from metadrive.component.lane.abs_lane import AbstractLane
from metadrive.component.pgblock.curve import Curve
from metadrive.component.pgblock.ramp import InRampOnStraight, OutRampOnStraight
from metadrive.component.pgblock.straight import Straight
from metadrive.component.road_network import Road
from metadrive.component.static_object.traffic_object import TrafficCone, TrafficWarning, TrafficBarrier
from metadrive.engine.engine_utils import get_engine
from metadrive.manager.base_manager import BaseManager


class TrafficObjectManager(BaseManager):
    """
    This class is used to manager all static object, such as traffic cones, warning tripod.
    """
    PRIORITY = 9

    # the distance between break-down vehicle and alert
    ALERT_DIST = 10

    # accident scene setting
    ACCIDENT_AREA_LEN = 10

    # distance between two cones
    CONE_LONGITUDE = 2
    CONE_LATERAL = 1
    PROHIBIT_SCENE_PROB = 0.67  # the reset is the probability of break_down_scene

    def __init__(self):
        super(TrafficObjectManager, self).__init__()
        self.accident_prob = 0.
        self.accident_lanes = []

    def before_reset(self):
        """
        Clear all objects in th scene
        """
        super(TrafficObjectManager, self).before_reset()
        self.accident_prob = self.engine.global_config["accident_prob"]

    def reset(self):
        """
        Generate an accident scene or construction scene on block
        :return: None
        """
        self.accident_lanes = []
        engine = get_engine()
        accident_prob = self.accident_prob
        if abs(accident_prob - 0.0) < 1e-2:
            return
        for block in engine.current_map.blocks:
            if type(block) not in [Straight, Curve, InRampOnStraight, OutRampOnStraight]:
                # blocks with exists do not generate accident scene
                continue
            if self.np_random.rand() > accident_prob:
                # prob filter
                continue

            road_1 = Road(block.pre_block_socket.positive_road.end_node, block.road_node(0, 0))
            road_2 = Road(block.road_node(0, 0), block.road_node(0, 1)) if not isinstance(block, Straight) else None

            if self.np_random.rand() > self.PROHIBIT_SCENE_PROB:
                accident_road = self.np_random.choice([road_1, road_2]) if not isinstance(block, Curve) else road_2
                accident_road = road_1 if accident_road is None else accident_road
                is_ramp = isinstance(block, InRampOnStraight) or isinstance(block, OutRampOnStraight)
                on_left = True if self.np_random.rand() > 0.5 or (accident_road is road_2 and is_ramp) else False
                accident_lane_idx = 0 if on_left else -1
                lane = accident_road.get_lanes(engine.current_map.road_network)[accident_lane_idx]
                longitude = lane.length - self.ACCIDENT_AREA_LEN - 5

                lateral_len = engine.current_map.config[engine.current_map.LANE_WIDTH]

                lane = engine.current_map.road_network.get_lane(accident_road.lane_index(accident_lane_idx))
                self.accident_lanes.append(accident_road.get_lanes(engine.current_map.road_network)[accident_lane_idx])
                self.prohibit_scene(lane, longitude, lateral_len, on_left)
            else:
                accident_road = self.np_random.choice([road_1, road_2])
                accident_road = road_1 if accident_road is None else accident_road
                is_ramp = isinstance(block, InRampOnStraight) or isinstance(block, OutRampOnStraight)
                on_left = True if self.np_random.rand() > 0.5 or (accident_road is road_2 and is_ramp) else False
                lanes = accident_road.get_lanes(engine.current_map.road_network)
                accident_lane_idx = self.np_random.randint(0, len(lanes) - 1) if on_left else -1
                lane = lanes[accident_lane_idx]
                self.accident_lanes.append(accident_road.get_lanes(engine.current_map.road_network)[accident_lane_idx])
                longitude = self.np_random.rand() * lane.length / 2 + lane.length / 2
                if self.np_random.rand() > 0.5:
                    self.break_down_scene(lane, longitude)
                else:
                    self.barrier_scene(lane, longitude)

    def break_down_scene(self, lane: AbstractLane, longitude: float):
        v_config = {"spawn_lane_index": lane.index, "spawn_longitude": float(longitude)}
        breakdown_vehicle = self.spawn_object(
            self.engine.traffic_manager.random_vehicle_type(), vehicle_config=v_config
        )
        breakdown_vehicle.set_break_down()
        longitude = longitude - self.ALERT_DIST
        lateral = 0
        self.spawn_object(
            TrafficWarning,
            lane=lane,
            position=lane.position(longitude, lateral),
            static=self.engine.global_config["static_traffic_object"],
            heading_theta=lane.heading_theta_at(longitude)
        )

    def barrier_scene(self, lane, longitude):
        longitude = longitude
        lateral = 0
        self.spawn_object(
            TrafficBarrier,
            lane=lane,
            position=lane.position(longitude, lateral),
            static=self.engine.global_config["static_traffic_object"],
            heading_theta=lane.heading_theta_at(longitude)
        )

    def prohibit_scene(self, lane: AbstractLane, longitude_position: float, lateral_len: float, on_left=False):
        """
        Generate an accident scene on the most left or most right lane
        :param lane object
        :param longitude_position: longitude position of the accident on the lane
        :param lateral_len: the distance that traffic cones extend on lateral direction
        :param on_left: on left or right side
        :return: None
        """
        lat_num = int(lateral_len / self.CONE_LATERAL)
        longitude_num = int(self.ACCIDENT_AREA_LEN / self.CONE_LONGITUDE)
        lat_1 = [lat * self.CONE_LATERAL for lat in range(lat_num)]
        lat_2 = [lat_num * self.CONE_LATERAL] * (longitude_num + 1)
        lat_3 = [(lat_num - lat - 1) * self.CONE_LATERAL for lat in range(int(lat_num))]

        total_long_num = lat_num * 2 + longitude_num + 1
        pos = [
            (long * self.CONE_LONGITUDE, lat - lane.width / 2)
            for long, lat in zip(range(-int(total_long_num / 2), int(total_long_num / 2)), lat_1 + lat_2 + lat_3)
        ]
        left = 1 if on_left else -1
        for p in pos:
            p_ = (p[0] + longitude_position, left * p[1])
            position = lane.position(p_[0], p_[1])
            heading_theta = lane.heading_theta_at(p_[0])
            self.spawn_object(
                TrafficCone,
                lane=lane,
                position=position,
                heading_theta=heading_theta,
                static=self.engine.global_config["static_traffic_object"]
            )

    def set_state(self, state: dict, old_name_to_current=None):
        """
        Copied from super(). Restoring some states before reassigning value to spawned_objets
        """
        assert self.episode_step == 0, "This func can only be called after env.reset() without any env.step() called"
        if old_name_to_current is None:
            old_name_to_current = {key: key for key in state.keys()}
        spawned_objects = state["spawned_objects"]
        ret = {}
        for name, class_name in spawned_objects.items():
            current_name = old_name_to_current[name]
            name_obj = self.engine.get_objects([current_name])
            assert current_name in name_obj and name_obj[current_name
                                                         ].class_name == class_name, "Can not restore mappings!"
            # Restore some internal states
            name_obj[current_name].lane = self.engine.current_map.road_network.get_lane(
                name_obj[current_name].lane.index
            )

            ret[current_name] = name_obj[current_name]
        self.spawned_objects = ret



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bpy
import itertools


def quat_to_list(quat):
    return [quat.x, quat.y, quat.z, quat.w]


def matrix_to_list(matrix):
    return list(itertools.chain(*map(tuple, matrix.col)))


def get_bone_matrix_local(bone):
    if isinstance(bone, bpy.types.PoseBone):
        return bone.matrix
    if isinstance(bone, bpy.types.EditBone):
        return bone.matrix
    elif isinstance(bone, bpy.types.Bone):
        return bone.matrix_local


def get_bone_matrix(bone, armature):
    bone_matrix = get_bone_matrix_local(bone)
    if bone.parent:
        parent_bone_matrix = get_bone_matrix_local(bone.parent)
        return parent_bone_matrix.inverted() @ bone_matrix
    else:  # root bone
        return bone_matrix


def get_object_matrix(obj, armature=None):
    if armature is None:
        return obj.matrix_local
    else:  # skinned mesh/object
        return (obj.matrix_world @ armature.matrix_world @ armature.matrix_world.inverted())


def get_inverse_bind_matrix(bone, armature):
    return bone.matrix_local.inverted() @ armature.matrix_world.inverted()



import colorsys
import math

from panda3d.core import CS_linear, PNMImage, PNMFileTypeRegistry


class Palette2LUT(object):
    def __init__(self):
        self.tile_width = 64  # cube width -> red-green tile width
        self.atlas_width = int(math.sqrt(self.tile_width))  # altas width in number of tiles
        lut_size = self.tile_width * self.atlas_width

        reg = PNMFileTypeRegistry.get_global_ptr()
        ftype = reg.get_type_from_extension('a.png')

        self.lut = PNMImage(lut_size, lut_size, 3, 255, ftype, CS_linear)

    def set_color(self, color_in, color_out):
        r, g, b = color_in

        tile_x = int(round(r * (self.tile_width - 1)))
        tile_y = int(round((1 - g) * (self.tile_width - 1)))

        tile_row = int(round(b * (self.tile_width - 1))) // self.atlas_width
        tile_col = int(round(b * (self.tile_width - 1))) % self.atlas_width

        x = tile_col * self.tile_width + tile_x
        y = tile_row * self.tile_width + tile_y

        r, g, b = color_out
        self.lut.set_xel(x, y, r, g, b)

    def convert(self, ipath):
        if ipath is None:
            for r in range(256):
                for g in range(256):
                    for b in range(256):
                        color = (r / 255, g / 255, b / 255)
                        self.set_color(color, color)
            return

        palette = PNMImage()
        palette.read(ipath)

        quality = 256
        for h in range(quality + 1):
            x = int(round(h / quality * (palette.get_read_x_size() - 1)))
            h0, s0, v0 = colorsys.rgb_to_hsv(*palette.get_xel(x, 0))

            for s in range(quality + 1):
                for v in range(quality + 1):
                    color_in = colorsys.hsv_to_rgb(h / quality, s / quality, v / quality)
                    color_out = colorsys.hsv_to_rgb(h0, s / quality, v / quality)
                    self.set_color(color_in, color_out)

    def save(self, opath):
        self.lut.write(opath)



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import math

from panda3d.core import CS_linear, PNMImage, PNMFileTypeRegistry, LMatrix4d
from panda3d.egg import (EggData, EggComment, EggGroup, EggPolygon, EggVertex, EggVertexPool, EggTexture)


class CardMaker(object):
    def __init__(self, animations_frames, images, empty=None, prefix=None, fps=30, scale=1, type='png'):
        self.animations = {}
        for i, frames in enumerate(animations_frames):
            self.animations[i] = frames
        self.images = images
        if empty:
            self.images.insert(0, empty)
        self.prefix = prefix or ''
        self.fps = fps
        self.scale = scale
        self.type = type
        self.empty = empty

        self.index_poly = 0
        self.index_vertex = 0

    def _get_frames_num(self):
        frames_total = sum(self.animations.values())
        if self.empty:
            frames_total += 1
        return frames_total

    def _add_polygon(self, name=None, group_main=None, egg_texture=None, egg_vertex_pool=None, is_empty=False):
        frames_total = self._get_frames_num()
        frames_rows = math.ceil(math.sqrt(frames_total))  # rows == columns

        # vertical mirror UV, because spritesheet goes from top to bottom
        y = (frames_rows - 1) - math.floor(self.index_poly / frames_rows)
        x = self.index_poly % frames_rows

        size_tile = 1 / frames_rows
        coords_uvs = (
            ((-1, 1, 0), (x * size_tile, (y + 1) * size_tile)),
            ((-1, -1, 0), (x * size_tile, y * size_tile)),
            ((1, -1, 0), ((x + 1) * size_tile, y * size_tile)),
            ((1, 1, 0), ((x + 1) * size_tile, (y + 1) * size_tile)),
        )
        if not name:
            self.empty_uvs = coords_uvs
            return

        if is_empty:
            coords_uvs = self.empty_uvs

        vertices = []
        for coord, uv in coords_uvs:
            vertex = EggVertex()
            vertex.set_pos(coord)
            vertex.set_uv(uv)

            egg_vertex_pool.add_vertex(vertex, self.index_vertex)
            self.index_vertex += 1

            vertices.append(vertex)

        poly = EggPolygon(name)
        poly.add_texture(egg_texture)
        for vertex in vertices:
            poly.add_vertex(vertex)

        group_poly = EggGroup(name)
        group_poly.add_child(poly)
        group_main.add_child(group_poly)

        self.index_poly += 1

    def _add_geom(self, name, count, egg_texture, egg_vertex_pool):
        egg_group = EggGroup(name)
        egg_group.set_switch_flag(True)  # enable animation
        egg_group.set_switch_fps(self.fps)  # animation fps

        for i in range(count):
            self._add_polygon('{}_frame_{:04d}'.format(name, self.index_poly), egg_group, egg_texture, egg_vertex_pool)

        if self.empty:
            self._add_polygon(
                '{}_frame_{:04d}'.format(name, self.index_poly), egg_group, egg_texture, egg_vertex_pool, is_empty=True
            )

        return egg_group

    def _make_spritesheet(self, path):
        reg = PNMFileTypeRegistry.get_global_ptr()
        ftype = reg.get_type_from_extension(os.path.basename(path))
        frames_total = self._get_frames_num()
        frames_rows = math.ceil(math.sqrt(frames_total))  # rows == columns
        frame_size = 0

        spritesheet = None

        i = 0
        while True:
            if i >= len(self.images):
                break
            ipath = self.images[i]

            image = PNMImage()
            image.read(ipath)

            if not frame_size:
                frame_size = image.get_read_x_size()

            if spritesheet is None:
                size = int(frames_rows * frame_size)
                spritesheet = PNMImage(size, size, 4, 255, ftype, CS_linear)
                spritesheet.add_alpha()

            y = i // frames_rows
            x = i % frames_rows

            spritesheet.blend_sub_image(
                image,
                x * frame_size,
                y * frame_size,  # (x, y) to
                0,
                0,  # (x, y) from
                frame_size,
                frame_size,
                1
            )

            i += 1

        if spritesheet is not None:
            spritesheet.write(path)

    def make(self, path):
        egg_root = EggData()

        egg_comment = EggComment(
            '', 'KITSUNETSUKI Asset Tools by kitsune.ONE - '
            'https://github.com/kitsune-ONE-team/KITSUNETSUKI-Asset-Tools'
        )

        spritesheet = path.replace('.egg', '.{}'.format(self.type))
        self._make_spritesheet(spritesheet)

        name = os.path.basename(spritesheet).rpartition('.')[0]
        egg_texture = EggTexture(name, '{}{}'.format(self.prefix, spritesheet))
        egg_vertex_pool = EggVertexPool('vpool')

        egg_root.add_child(egg_comment)
        egg_root.add_child(egg_texture)
        egg_root.add_child(egg_vertex_pool)

        if self.empty:
            self._add_polygon()

        for animation, frames in self.animations.items():
            egg_group = self._add_geom('animation_{:04d}'.format(animation), frames, egg_texture, egg_vertex_pool)
            egg_group.add_matrix4(LMatrix4d.scale_mat(self.scale))
            egg_root.add_child(egg_group)

        # print(egg_root)
        egg_root.write_egg(path)



import math
import queue
from collections import deque
from typing import Tuple

import numpy as np
from direct.controls.InputState import InputState
from panda3d.core import Vec3, Point3, PNMImage
from panda3d.core import WindowProperties

from metadrive.constants import CollisionGroup
from metadrive.engine.engine_utils import get_engine
from metadrive.utils.coordinates_shift import panda_heading, panda_vector
from metadrive.utils.cuda import check_cudart_err

_cuda_enable = True
try:
    import cupy as cp
    from OpenGL.GL import GL_TEXTURE_2D  # noqa F403
    from cuda import cudart
    from cuda.cudart import cudaGraphicsRegisterFlags
    from panda3d.core import GraphicsOutput, Texture, GraphicsStateGuardianBase, DisplayRegionDrawCallbackData
except ImportError:
    _cuda_enable = False


class MainCamera:
    """
    Only chase vehicle now
    """

    queue_length = 3
    CHASE_TASK_NAME = "update main chase camera"
    TOP_DOWN_TASK_NAME = "update main bird camera"
    FOLLOW_LANE = False
    TOP_DOWN_VIEW_HEIGHT = 120
    WHEEL_SCROLL_SPEED = 10
    MOUSE_RECOVER_TIME = 8
    STATIC_MOUSE_HOLD_TIME = 100  # in steps
    MOUSE_MOVE_INTO_LATENCY = 2
    MOUSE_SPEED_MULTIPLIER = 1

    def __init__(self, engine, camera_height: float, camera_dist: float):
        self._origin_height = camera_height
        # self.engine = engine

        # vehicle chase camera
        self.camera = engine.camera
        self.camera_queue = None
        self.camera_dist = camera_dist
        self.camera_pitch = -engine.global_config["camera_pitch"] if engine.global_config["camera_pitch"
                                                                                          ] is not None else None
        self.camera_smooth = engine.global_config["camera_smooth"]
        self.direction_running_mean = deque(maxlen=20 if self.camera_smooth else 1)
        self.world_light = engine.world_light  # light chases the chase camera, when not using global light
        self.inputs = InputState()
        self.current_track_vehicle = None

        # height control
        self.chase_camera_height = camera_height
        self.inputs.watchWithModifiers('high', '+')
        self.inputs.watchWithModifiers('high', '=')
        self.inputs.watchWithModifiers('low', '-')
        self.inputs.watchWithModifiers('low', '_')

        # free bird view camera
        self.top_down_camera_height = self.TOP_DOWN_VIEW_HEIGHT
        self.camera_x = 0
        self.camera_y = 0
        self.inputs.watchWithModifiers('up', 'w')
        self.inputs.watchWithModifiers('down', 's')
        self.inputs.watchWithModifiers('left', 'a')
        self.inputs.watchWithModifiers('right', 'd')
        self.inputs.watchWithModifiers('left_rotate', '[')
        self.inputs.watchWithModifiers('right_rotate', ']')
        self.engine = engine

        engine.accept("wheel_up", self._wheel_up_height)
        engine.accept("wheel_down", self._wheel_down_height)
        engine.accept("mouse1", self._move_to_pointer)

        # default top-down
        self.top_down_camera_height = engine.global_config["top_down_camera_initial_z"]
        self.camera_x = engine.global_config["top_down_camera_initial_x"]
        self.camera_y = engine.global_config["top_down_camera_initial_y"]
        self.camera_rotate = 0
        engine.interface.stop_track()
        engine.task_manager.add(self._top_down_task, self.TOP_DOWN_TASK_NAME, extraArgs=[], appendTask=True)

        # TPP rotate
        if not engine.global_config["show_mouse"]:
            props = WindowProperties()
            props.setCursorHidden(True)
            props.setMouseMode(WindowProperties.MConfined)
            engine.win.requestProperties(props)
        self.mouse_rotate = 0
        self.last_mouse_pos = engine.mouseWatcherNode.getMouseX() if self.has_mouse else 0
        self.static_timer = 0
        self.move_into_window_timer = 0
        self._in_recover = False
        self._last_frame_has_mouse = False

        need_cuda = engine.global_config["vehicle_config"]["image_source"] == "main_camera"
        self.enable_cuda = engine.global_config["image_on_cuda"] and need_cuda

        self.cuda_graphics_resource = None
        self.engine.sensors["main_camera"] = self
        if self.enable_cuda:
            assert _cuda_enable, "Can not enable cuda rendering pipeline"

            # returned tensor property
            self.cuda_dtype = np.uint8
            self.cuda_shape = engine.global_config["window_size"]
            self.cuda_strides = None
            self.cuda_order = "C"

            self._cuda_buffer = None

            # make texture
            self.cuda_texture = Texture()
            engine.win.addRenderTexture(self.cuda_texture, GraphicsOutput.RTMCopyTexture)

            def _callback_func(cbdata: DisplayRegionDrawCallbackData):
                # print("DRAW CALLBACK!!!!!!!!!!!!!!!11")
                cbdata.upcall()
                if not self.registered and self.texture_context_future.done():
                    self.register()
                if self.registered:
                    with self as array:
                        self.cuda_rendered_result = array

            # Fill the buffer due to multi-thread (3 stage)
            for _ in range(3):
                engine.graphicsEngine.renderFrame()
            engine.cam.node().getDisplayRegion(0).setDrawCallback(_callback_func)

            self.gsg = GraphicsStateGuardianBase.getDefaultGsg()
            self.texture_context_future = self.cuda_texture.prepare(self.gsg.prepared_objects)
            self.cuda_texture_identifier = None
            self.new_cuda_mem_ptr = None
            self.cuda_rendered_result = None

    def set_bird_view_pos(self, position):
        if self.engine.task_manager.hasTaskNamed(self.TOP_DOWN_TASK_NAME):
            # adjust hpr
            p_pos = panda_vector(position)
            self.camera_x, self.camera_y = p_pos[0], p_pos[1]
            self.camera_rotate = 0
            self.engine.task_manager.add(self._top_down_task, self.TOP_DOWN_TASK_NAME, extraArgs=[], appendTask=True)

    def reset(self):
        self.direction_running_mean.clear()

    def update_mouse_info(self):
        self.move_into_window_timer -= 1 if self.move_into_window_timer > 0 else 0
        if self.mouse_into_window:
            self._in_recover = True
            self.set_mouse_to_center()
            self.move_into_window_timer = self.MOUSE_MOVE_INTO_LATENCY

        if not self._in_recover and self.has_mouse and not self.mouse_into_window and self.move_into_window_timer == 0:
            new_mouse_pos = self.engine.mouseWatcherNode.getMouseX()
            last_rotate = self.mouse_rotate
            self.mouse_rotate = -new_mouse_pos * self.MOUSE_SPEED_MULTIPLIER
            diff = abs(last_rotate - self.mouse_rotate)
            if diff == 0.:
                self.static_timer += 1
            else:
                self.static_timer = 0
            if self.static_timer > self.STATIC_MOUSE_HOLD_TIME:
                self._in_recover = True
                self.set_mouse_to_center()
        else:
            self.mouse_rotate += -self.mouse_rotate / self.MOUSE_RECOVER_TIME

        if self._in_recover and abs(self.mouse_rotate) < 0.01:
            self._in_recover = False
            self.static_timer = 0
            self.last_mouse_pos = 0

        self._last_frame_has_mouse = self.has_mouse

    def _chase_task(self, vehicle, task):
        self.update_mouse_info()
        self.chase_camera_height = self._update_height(self.chase_camera_height)
        chassis_pos = vehicle.chassis.get_pos()
        self.camera_queue.put(chassis_pos)
        if not self.FOLLOW_LANE:
            forward_dir = vehicle.system.get_forward_vector()
            # camera is facing to y
            current_forward_dir = [forward_dir[0], forward_dir[1]]
        else:
            current_forward_dir = self._dir_of_lane(vehicle.navigation.current_ref_lanes[0], vehicle.position)
        self.direction_running_mean.append(current_forward_dir)
        forward_dir = np.mean(self.direction_running_mean, axis=0) if self.camera_smooth else current_forward_dir
        forward_dir[0] = np.cos(self.mouse_rotate) * current_forward_dir[0] - np.sin(self.mouse_rotate) * \
                         current_forward_dir[1]
        forward_dir[1] = np.sin(self.mouse_rotate) * current_forward_dir[0] + np.cos(self.mouse_rotate) * \
                         current_forward_dir[1]

        # don't put this line to if-else, strange bug happened
        camera_pos = list(self.camera_queue.get())
        if not self.camera_smooth:
            camera_pos = chassis_pos
        camera_pos[2] += self.chase_camera_height + vehicle.HEIGHT / 2
        camera_pos[0] += -forward_dir[0] * self.camera_dist
        camera_pos[1] += -forward_dir[1] * self.camera_dist

        self.camera.setPos(*camera_pos)
        if self.engine.global_config["show_coordinates"]:
            self.engine.set_coordinates_indicator_pos([chassis_pos[0], chassis_pos[1]])
        current_pos = vehicle.chassis.getPos()
        current_pos[2] += 2

        if self.camera_pitch is None:
            self.camera.lookAt(current_pos)
            # camera is facing to y
            self.camera.setH(vehicle.origin.getH() + np.rad2deg(self.mouse_rotate))
        else:
            # camera is facing to y
            self.camera.setH(vehicle.origin.getH())
            self.camera.setP(self.camera_pitch)
        if self.FOLLOW_LANE:
            self.camera.setH(
                self._heading_of_lane(vehicle.navigation.current_ref_lanes[0], vehicle.position) / np.pi * 180 - 90
            )

        if self.world_light is not None:
            self.world_light.step(current_pos)

        # Don't use reparentTo()
        # pos = vehicle.convert_to_world_coordinates([0.8, 0.], vehicle.position)
        # look_at = vehicle.convert_to_world_coordinates([10.4, 0.], vehicle.position)
        # self.camera.setPos(pos[0], pos[1], 1.5 + vehicle.origin.getZ())
        # self.camera.lookAt(look_at[0], look_at[1], 1.6 + vehicle.origin.getZ())
        return task.cont

    @staticmethod
    def _heading_of_lane(lane, pos: Tuple) -> float:
        """
        Calculate the heading of a position on lane
        :param lane: Abstract lane
        :param pos: Tuple, MetaDrive coordinates
        :return: heading theta
        """
        heading_theta = panda_heading(lane.heading_theta_at(lane.local_coordinates(pos)[0]))
        return heading_theta

    @staticmethod
    def _dir_of_lane(lane, pos: Tuple) -> Tuple:
        """
        Get direction of lane
        :param lane: AbstractLane
        :param pos: metadrive position, tuple
        :return: dir, tuple
        """
        heading = MainCamera._heading_of_lane(lane, pos)
        return math.cos(heading), math.sin(heading)

    def track(self, vehicle):
        """
        Use this function to chase a new vehicle !
        :param vehicle: Vehicle to chase
        :return: None
        """
        self.current_track_vehicle = vehicle
        self.engine.interface.track(vehicle)
        pos = None
        if self.FOLLOW_LANE:
            pos = self._pos_on_lane(vehicle)  # Return None if routing system is not ready
        if pos is None:
            pos = vehicle.position

        if self.engine.task_manager.hasTaskNamed(self.CHASE_TASK_NAME):
            self.engine.task_manager.remove(self.CHASE_TASK_NAME)
        if self.engine.task_manager.hasTaskNamed(self.TOP_DOWN_TASK_NAME):
            self.engine.task_manager.remove(self.TOP_DOWN_TASK_NAME)
        self.mouse_rotate = 0
        self.last_mouse_pos = self.engine.mouseWatcherNode.getMouseX() if self.has_mouse else 0
        self.static_timer = 0
        self.set_mouse_to_center()
        self.engine.task_manager.add(self._chase_task, self.CHASE_TASK_NAME, extraArgs=[vehicle], appendTask=True)
        self.camera_queue = queue.Queue(self.queue_length)
        for i in range(self.queue_length - 1):
            self.camera_queue.put(Vec3(pos[0], -pos[1], 0))

    @staticmethod
    def _pos_on_lane(vehicle) -> Tuple:
        """
        Recalculate cam place
        :param vehicle: BaseVehicle
        :return: position on the center lane
        """
        if vehicle.navigation.current_ref_lanes is None:
            raise ValueError("No routing module, I don't know which lane to follow")

        lane = vehicle.navigation.current_ref_lanes[0]
        lane_num = len(vehicle.navigation.current_ref_lanes)

        longitude, _ = lane.local_coordinates(vehicle.position)
        lateral = lane_num * lane.width / 2 - lane.width / 2
        return longitude, lateral

    def set_follow_lane(self, flag: bool):
        """
        Camera will go follow reference lane instead of vehicle
        :return: None
        """
        self.FOLLOW_LANE = flag

    def destroy(self):
        engine = get_engine()
        if engine.task_manager.hasTaskNamed(self.CHASE_TASK_NAME):
            engine.task_manager.remove(self.CHASE_TASK_NAME)
        if engine.task_manager.hasTaskNamed(self.TOP_DOWN_TASK_NAME):
            engine.task_manager.remove(self.TOP_DOWN_TASK_NAME)
        self.current_track_vehicle = None
        if self.registered:
            self.unregister()
            self.camera.node().getDisplayRegion(0).clearDrawCallback()

    def stop_track(self, bird_view_on_current_position=True):
        self.engine.interface.stop_track()
        if self.engine.task_manager.hasTaskNamed(self.CHASE_TASK_NAME):
            self.engine.task_manager.remove(self.CHASE_TASK_NAME)
        if not self.engine.task_manager.hasTaskNamed(self.TOP_DOWN_TASK_NAME):
            # adjust hpr
            if bird_view_on_current_position:
                current_pos = self.camera.getPos()
                self.camera_x, self.camera_y = current_pos[0], current_pos[1]
                self.camera_rotate = 0
            self.engine.task_manager.add(self._top_down_task, self.TOP_DOWN_TASK_NAME, extraArgs=[], appendTask=True)

    def _top_down_task(self, task):
        self.top_down_camera_height = self._update_height(self.top_down_camera_height)

        if self.inputs.isSet("up"):
            self.camera_y += 1.0
        if self.inputs.isSet("down"):
            self.camera_y -= 1.0
        if self.inputs.isSet("left"):
            self.camera_x -= 1.0
        if self.inputs.isSet("right"):
            self.camera_x += 1.0

        self.camera.setPos(self.camera_x, self.camera_y, self.top_down_camera_height)
        if self.engine.global_config["show_coordinates"]:
            self.engine.set_coordinates_indicator_pos([self.camera_x, self.camera_y])
        self.camera.lookAt(self.camera_x, self.camera_y, 0)

        if self.inputs.isSet("right_rotate"):
            self.camera_rotate += 3
        if self.inputs.isSet("left_rotate"):
            self.camera_rotate -= 3
        self.camera.setH(self.camera_rotate)
        return task.cont

    def _update_height(self, height):
        if self.inputs.isSet("high"):
            height += 1.0
        if self.inputs.isSet("low"):
            height -= 1.0
        return height

    def _wheel_down_height(self):
        if self.engine.task_manager.hasTaskNamed(self.TOP_DOWN_TASK_NAME):
            self.top_down_camera_height += self.WHEEL_SCROLL_SPEED
        else:
            self.chase_camera_height += self.WHEEL_SCROLL_SPEED

    def _wheel_up_height(self):
        if self.engine.task_manager.hasTaskNamed(self.TOP_DOWN_TASK_NAME):
            self.top_down_camera_height -= self.WHEEL_SCROLL_SPEED
        else:
            self.chase_camera_height -= self.WHEEL_SCROLL_SPEED

    def _move_to_pointer(self):
        if self.engine.task_manager.hasTaskNamed(self.TOP_DOWN_TASK_NAME):
            # Get to and from pos in camera coordinates
            pMouse = self.engine.mouseWatcherNode.getMouse()
            pFrom = Point3()
            pTo = Point3()
            self.engine.cam.node().getLens().extrude(pMouse, pFrom, pTo)

            # Transform to global coordinates
            pFrom = self.engine.render.getRelativePoint(self.engine.cam, pFrom)
            pTo = self.engine.render.getRelativePoint(self.engine.cam, pTo)
            ret = self.engine.physics_world.dynamic_world.rayTestClosest(pFrom, pTo, CollisionGroup.Terrain)
            self.camera_x = ret.getHitPos()[0]
            self.camera_y = ret.getHitPos()[1]

    def is_bird_view_camera(self):
        return True if self.engine.task_manager.hasTaskNamed(self.TOP_DOWN_TASK_NAME) else False

    @property
    def has_mouse(self):
        if self.engine.mouseWatcherNode is None:
            return False
        return True if self.engine.mouseWatcherNode.hasMouse() else False

    def set_mouse_to_center(self):
        mouse_id = 0
        if self.has_mouse:
            win_middle_x = self.engine.win.getXSize() / 2
            win_middle_y = self.engine.win.getYSize() / 2
            self.engine.win.movePointer(mouse_id, int(win_middle_x), int(win_middle_y))

    @property
    def mouse_into_window(self):
        return True if not self._last_frame_has_mouse and self.has_mouse else False

    def get_pixels_array(self, vehicle, clip):
        engine = get_engine()
        assert engine.main_camera.current_track_vehicle is vehicle, "Tracked vehicle mismatch"
        if self.enable_cuda:
            assert self.cuda_rendered_result is not None
            img = self.cuda_rendered_result[..., :-1][..., ::-1][::-1]
        else:
            origin_img = engine.win.getDisplayRegion(1).getScreenshot()
            img = np.frombuffer(origin_img.getRamImage().getData(), dtype=np.uint8)
            img = img.reshape((origin_img.getYSize(), origin_img.getXSize(), 4))
            img = img[::-1]
            img = img[..., :-1]

        if not clip:
            return img.astype(np.uint8)
        else:
            return img / 255

    def __del__(self):
        if self.enable_cuda:
            self.unregister()

    """
    Following functions are cuda support
    """

    @property
    def cuda_array(self):
        assert self.mapped
        return cp.ndarray(
            shape=(self.cuda_shape[1], self.cuda_shape[0], 4),
            dtype=self.cuda_dtype,
            strides=self.cuda_strides,
            order=self.cuda_order,
            memptr=self._cuda_buffer
        )

    @property
    def cuda_buffer(self):
        assert self.mapped
        return self._cuda_buffer

    @property
    def graphics_resource(self):
        assert self.registered
        return self.cuda_graphics_resource

    @property
    def registered(self):
        return self.cuda_graphics_resource is not None

    @property
    def mapped(self):
        return self._cuda_buffer is not None

    def __enter__(self):
        return self.map()

    def __exit__(self, exc_type, exc_value, trace):
        self.unmap()
        return False

    def register(self):
        self.cuda_texture_identifier = self.texture_context_future.result().getNativeId()
        assert self.cuda_texture_identifier is not None
        if self.registered:
            return self.cuda_graphics_resource
        self.cuda_graphics_resource = check_cudart_err(
            cudart.cudaGraphicsGLRegisterImage(
                self.cuda_texture_identifier, GL_TEXTURE_2D, cudaGraphicsRegisterFlags.cudaGraphicsRegisterFlagsReadOnly
            )
        )
        return self.cuda_graphics_resource

    def unregister(self):
        if self.registered:
            self.unmap()
            self.cuda_graphics_resource = check_cudart_err(
                cudart.cudaGraphicsUnregisterResource(self.cuda_graphics_resource)
            )

    def map(self, stream=0):
        if not self.registered:
            raise RuntimeError("Cannot map an unregistered buffer.")
        if self.mapped:
            return self._cuda_buffer
        check_cudart_err(cudart.cudaGraphicsMapResources(1, self.cuda_graphics_resource, stream))
        array = check_cudart_err(cudart.cudaGraphicsSubResourceGetMappedArray(self.graphics_resource, 0, 0))
        channelformat, cudaextent, flag = check_cudart_err(cudart.cudaArrayGetInfo(array))

        depth = 1
        byte = 4  # four channel
        if self.new_cuda_mem_ptr is None:
            success, self.new_cuda_mem_ptr = cudart.cudaMalloc(cudaextent.height * cudaextent.width * byte * depth)
        check_cudart_err(
            cudart.cudaMemcpy2DFromArray(
                self.new_cuda_mem_ptr, cudaextent.width * byte * depth, array, 0, 0, cudaextent.width * byte * depth,
                cudaextent.height, cudart.cudaMemcpyKind.cudaMemcpyDeviceToDevice
            )
        )
        if self._cuda_buffer is None:
            self._cuda_buffer = cp.cuda.MemoryPointer(
                cp.cuda.UnownedMemory(self.new_cuda_mem_ptr, cudaextent.width * depth * byte * cudaextent.height, self),
                0
            )
        return self.cuda_array

    def unmap(self, stream=None):
        if not self.registered:
            raise RuntimeError("Cannot unmap an unregistered buffer.")
        if not self.mapped:
            return self
        self._cuda_buffer = check_cudart_err(cudart.cudaGraphicsUnmapResources(1, self.cuda_graphics_resource, stream))
        return self

    def get_image(self):
        # The Tracked obj arg is only for compatibility
        img = PNMImage()
        self.engine.win.getScreenshot(img)
        return img

    def save_image(self, tracked_obj, file_name="main_camera.png", **kwargs):
        # The Tracked obj arg is only for compatibility
        img = self.get_image()
        img.write(file_name)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import socket
from threading import Thread

from metadrive.render_pipeline.rpcore.rpobject import RPObject


class NetworkCommunication(RPObject):
    """ Listener which accepts messages on several ports to detect incoming updates.
    Also provides functionality to send updates. """

    CONFIG_PORT = 63324
    DAYTIME_PORT = 63325
    MATERIAL_PORT = 63326

    @classmethod
    def send_async(cls, port, message):
        """ Starts a new thread which sends a given message to a port """
        thread = Thread(target=cls.__send_message_async, args=(port, message), name="NC-SendAsync")
        thread.setDaemon(True)
        thread.start()
        return thread

    @classmethod
    def listen_threaded(cls, port, callback):
        """ Starts a new thread listening to the given port """
        thread = Thread(target=cls.__listen_forever, args=(port, callback), name="NC-ListenForever")
        thread.setDaemon(True)
        thread.start()
        return thread

    @staticmethod
    def __send_message_async(port, message=""):
        """ Sends a given message to a given port and immediately returns. """
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            sock.sendto(message.encode("utf-8"), ("127.0.0.1", port))
        finally:
            sock.close()

    @staticmethod
    def __listen_forever(port, callback):
        """ Listens to a given port, and calls callback in case a message
        arrives. This method never returns, except when the connection closed or
        could not be established. """
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            sock.bind(("127.0.0.1", port))
            while True:
                data, addr = sock.recvfrom(1024)  # pylint: disable=unused-variable
                callback(data.decode("utf-8"))
        finally:
            sock.close()

    def __init__(self, pipeline):
        """ Creates the listener service. This also starts listening on the various
        ports for updates """
        RPObject.__init__(self)
        self._pipeline = pipeline
        self._config_updates = set()
        self._daytime_updates = set()
        self._material_updates = set()
        self._config_thread = self.listen_threaded(self.CONFIG_PORT, self._config_updates.add)
        self._daytime_thread = self.listen_threaded(self.DAYTIME_PORT, self._daytime_updates.add)
        self._material_thread = self.listen_threaded(self.MATERIAL_PORT, self._material_updates.add)

    def update(self):
        """ Update task which gets called every frame and executes the changes.
        This takes the incoming scheduled commands and processes one at a time."""
        while self._config_updates:
            cmd = self._config_updates.pop()
            self._handle_config_command(cmd)
        while self._daytime_updates:
            cmd = self._daytime_updates.pop()
            self._handle_daytime_command(cmd)
        while self._material_updates:
            cmd = self._material_updates.pop()
            self._handle_material_command(cmd)

    def _handle_daytime_command(self, cmd):
        """ Handles a daytime command. This could either be a command to set
        the time, or a command to reload the time of day configuration. """
        if cmd.startswith("settime "):
            daytime = float(cmd.split()[1])
            self._pipeline.daytime_mgr.time = daytime
        elif cmd.startswith("loadconf"):
            self._pipeline.plugin_mgr.load_daytime_overrides("/$$rpconfig/daytime.yaml")
        else:
            self.warn("Recieved unkown daytime command:", cmd)

    def _handle_config_command(self, cmd):
        """ Handles an incomming configuration command. Currently this can only
        be an update of a plugin setting """
        if cmd.startswith("setval "):
            parts = cmd.split()
            setting_parts = parts[1].split(".")
            self._pipeline.plugin_mgr.on_setting_changed(setting_parts[0], setting_parts[1], parts[2])
        else:
            self.warn("Recieved unkown plugin command:", cmd)

    def _handle_material_command(self, cmd):
        """ Handles an incomming material command """
        if cmd.startswith("dump_materials"):
            path = cmd[len("dump_materials "):].strip()
            self.debug("Writing materials to", path)
            self._pipeline.export_materials(path)

        elif cmd.startswith("update_material"):

            data = cmd[len("update_material "):].strip()
            parts = data.split()
            self._pipeline.update_serialized_material(parts)

        else:
            self.warn("Recieved unkown plugin command:", cmd)



from __future__ import print_function

import os
import sys
import struct
from panda3d.core import Vec3, load_prc_file_data, Texture, GeomEnums, Vec2
from panda3d.core import OmniBoundingVolume
from direct.showbase.ShowBase import ShowBase

# Change to the current directory
os.chdir(os.path.dirname(os.path.realpath(__file__)))

# Insert the pipeline path to the system path, this is required to be
# able to import the pipeline classes
pipeline_path = "../../"

# Just a special case for my development setup, so I don't accidentally
# commit a wrong path. You can remove this in your own programs.
if not os.path.isfile(os.path.join(pipeline_path, "setup.py")):
    pipeline_path = "../../RenderPipeline/"

sys.path.insert(0, pipeline_path)

# Import the render pipeline class
from metadrive.render_pipeline.rpcore import RenderPipeline

from metadrive.render_pipeline.rpcore.water.projected_water import ProjectedWater

# This is a helper class for better camera movement - see below.
from metadrive.render_pipeline.rpcore.util.movement_controller import MovementController


class WaterOptions:
    """ This class stores all options required for the WaterDisplacement """
    size = 128

    wind_dir = Vec2(0.8, 0.6)
    wind_speed = 1600.0
    wind_dependency = 0.2
    choppy_scale = 1.3
    patch_length = 2000.0
    wave_amplitude = 0.35

    time_scale = 0.8
    normalization_factor = 150.0


class Application(ShowBase):
    def __init__(self):
        # Setup window size and title
        load_prc_file_data(
            "", """
            # win-size 1600 900
            window-title Render Pipeline - Projected Water Example
        """
        )

        # Construct the render pipeline
        self.render_pipeline = RenderPipeline()
        self.render_pipeline.create(self)
        self.render_pipeline.daytime_mgr.time = "8:17"
        # self.render_pipeline.daytime_mgr.time = "12:00"

        # Initialize the projected water
        self.water = ProjectedWater(WaterOptions)
        self.water.setup_water(pipeline=self.render_pipeline, water_level=-3.0)

        # Initialize movement controller, this is a convenience class
        # to provide an improved camera control compared to Panda3Ds default
        # mouse controller.
        self.controller = MovementController(self)
        self.controller.set_initial_position_hpr(Vec3(-23.2, -32.5, 5.3), Vec3(-33.8, -8.3, 0.0))
        self.controller.setup()


Application().run()



import copy
from metadrive.scenario.scenario_description import ScenarioDescription as SD
import logging
from typing import List, Tuple, Dict

from metadrive.component.lane.abs_lane import AbstractLane
from metadrive.component.road_network.base_road_network import BaseRoadNetwork
from metadrive.component.road_network.road import Road
from metadrive.constants import Decoration
from metadrive.utils.math import get_boxes_bounding_box
from metadrive.utils.pg.utils import get_lanes_bounding_box

logger = logging.getLogger(__name__)

LaneIndex = Tuple[str, str, int]
Route = List[LaneIndex]


class GraphLookupTable:
    def __init__(self, graph, debug):
        self.graph = graph
        self.debug = debug

    def get(self, position, return_all):
        log = dict()
        count = 0
        for _, (_from, to_dict) in enumerate(self.graph.items()):
            if _from == "decoration":
                continue
            for lanes_id, lanes in to_dict.items():
                lane = next(iter(lanes))
                log[count] = (lane.distance(position), (_from, lanes_id))
                count += 1

        distance_index_mapping = []
        for rank, candidate_count in enumerate(sorted(log, key=lambda key: log[key][0])):
            first_lane_distance, (section_id, lanes_id) = log[candidate_count]
            lanes = self.graph[section_id][lanes_id]
            for lane_id, lane in enumerate(lanes):
                if lanes_id == Decoration.start:
                    continue
                if lane_id == 0:
                    dist = first_lane_distance
                else:
                    dist = lane.distance(position)
                distance_index_mapping.append((dist, (section_id, lanes_id, lane_id)))
            # if rank > 10:
            #     # Take first rank 5 lanes into consideration. The number may related to the number of
            #     # lanes in intersection. We have 3 lanes in intersection, so computing the first 4 ranks can make
            #     # thing work. We choose take first 5 lanes into consideration.
            #     # In futurem we shall refactor the whole system, so this vulnerable code would be removed.
            #     break
        if self.graph.get(Decoration.start, False):
            for id, lane in enumerate(self.graph[Decoration.start][Decoration.end]):
                dist = lane.distance(position)
                distance_index_mapping.append((dist, (Decoration.start, Decoration.end, id)))

        distance_index_mapping = sorted(distance_index_mapping, key=lambda d: d[0])
        if return_all:
            return distance_index_mapping
        else:
            ret_ind = 0
            index = distance_index_mapping[ret_ind][1]
            distance = distance_index_mapping[ret_ind][0]
            return index, distance


class NodeRoadNetwork(BaseRoadNetwork):
    """
    This network uses two node to describe the road network graph, and the edges between two nodes represent road, which
    is a list of lanes connecting two lanes
    """
    graph: Dict[str, Dict[str, List[AbstractLane]]]

    def __init__(self, debug=False):
        super(NodeRoadNetwork, self).__init__()
        self.graph = {}
        self.indices = []
        self._graph_helper = None
        self.debug = debug
        self.is_initialized = False

    def after_init(self):
        assert not self.is_initialized
        self._update_indices()
        self._init_graph_helper()
        self.is_initialized = True

    def add(self, other, no_intersect=True):
        assert not self.is_initialized, "Adding new blocks should be done before road network initialization!"
        set_1 = set(self.graph) - {Decoration.start, Decoration.end}
        set_2 = set(other.graph) - {Decoration.start, Decoration.end}
        intersect = set_1.intersection(set_2)
        if len(intersect) != 0 and no_intersect:
            raise ValueError("Same start node {} in two road network".format(intersect))
        # handle decoration_lanes
        dec_lanes = self.get_all_decoration_lanes() + other.get_all_decoration_lanes()

        # PZH: Note, do not use deepcopy here! We wish to maintain the reference to Lane!
        # self.graph.update(copy.deepcopy(other.graph))
        self.graph.update(copy.copy(other.graph))

        self.update_decoration_lanes(dec_lanes)
        return self

    def __isub__(self, other):
        intersection = self.graph.keys() & other.graph.keys() - {Decoration.start, Decoration.end}
        if len(intersection) != 0:
            for k in intersection:
                self.graph.pop(k, None)
        if Decoration.start in other.graph.keys():
            for lane in other.graph[Decoration.start][Decoration.end]:
                if lane in self.graph[Decoration.start][Decoration.end]:
                    self.graph[Decoration.start][Decoration.end].remove(lane)
        return self

    def get_all_decoration_lanes(self) -> List:
        if Decoration.start in self.graph:
            return self.graph[Decoration.start][Decoration.end]
        else:
            return []

    def update_decoration_lanes(self, lanes):
        if len(lanes) == 0:
            return
        if Decoration.start in self.graph:
            self.graph.pop(Decoration.start, None)
        self.graph[Decoration.start] = {Decoration.end: lanes}

    def clear(self):
        self.graph.clear()

    def get_positive_lanes(self):
        """
        In order to remain the lane index, ret is a 2-dim array structure like [Road_lanes[lane_1, lane_2]]
        """
        ret = []
        for _from, _to_dict in self.graph.items():
            for _to, lanes in _to_dict.items():
                road = Road(_from, _to)
                if not road.is_negative_road() and road.is_valid_road():
                    ret.append(lanes)
        return ret

    def get_negative_lanes(self):
        """
        In order to remain the lane index, ret is a 2-dim array structure like like [Road_lanes[lane_1, lane_2]]
        """
        ret = []
        for _from, _to_dict in self.graph.items():
            for _to, lanes in _to_dict.items():
                road = Road(_from, _to)
                if road.is_negative_road() and road.is_valid_road():
                    ret.append(lanes)
        return ret

    def _get_bounding_box(self):
        """
        By using this bounding box, the edge length of x, y direction and the center of this road network can be
        easily calculated.
        :return: minimum x value, maximum x value, minimum y value, maximum y value
        """
        boxes = []
        for _from, to_dict in self.graph.items():
            for _to, lanes in to_dict.items():
                if len(lanes) == 0:
                    continue
                boxes.append(get_lanes_bounding_box(lanes))
        res_x_max, res_x_min, res_y_max, res_y_min = get_boxes_bounding_box(boxes)
        return res_x_min, res_x_max, res_y_min, res_y_max

    def remove_all_roads(self, start_node: str, end_node: str):
        """
        Remove all road between two road nodes
        :param start_node: start node name
        :param end_node: end node name
        :return: roads removed
        """
        ret = []
        paths = self.bfs_paths(start_node, end_node)
        for path in paths:
            for next_idx, node in enumerate(path[:-1], 1):
                road_removed = self.remove_road(Road(node, path[next_idx]))
                ret += road_removed
        return ret

    def remove_road(self, road):
        assert isinstance(road, Road), "Only Road Type can be deleted"
        ret = self.graph[road.start_node].pop(road.end_node)
        if len(self.graph[road.start_node]) == 0:
            self.graph.pop(road.start_node)
        return ret

    def add_road(self, road, lanes: List):
        assert isinstance(road, Road), "Only Road Type can be added to road network"
        if road.start_node not in self.graph:
            self.graph[road.start_node] = {}
        if road.end_node not in self.graph[road.start_node]:
            self.graph[road.start_node][road.end_node] = []
        self.graph[road.start_node][road.end_node] += lanes

    def add_lane(self, _from: str, _to: str, lane: AbstractLane) -> None:
        """
        A lane is encoded as an edge in the road network.

        :param _from: the node at which the lane starts.
        :param _to: the node at which the lane ends.
        :param AbstractLane lane: the lane geometry.
        """
        if _from not in self.graph:
            self.graph[_from] = {}
        if _to not in self.graph[_from]:
            self.graph[_from][_to] = []
        self.graph[_from][_to].append(lane)

    def _init_graph_helper(self):
        self._graph_helper = GraphLookupTable(self.graph, self.debug)

    def get_lane(self, index: LaneIndex) -> AbstractLane:
        """
        Get the lane geometry corresponding to a given index in the road network.

        :param index: a tuple (origin node, destination node, lane id on the road).
        :return: the corresponding lane geometry.
        """
        _from, _to, _id = index
        if _id is None and len(self.graph[_from][_to]) == 1:
            _id = 0
        return self.graph[_from][_to][_id]

    def _update_indices(self):
        indexes = []
        for _from, to_dict in self.graph.items():
            for _to, lanes in to_dict.items():
                for _id, l in enumerate(lanes):
                    indexes.append((_from, _to, _id))
        self.indices = indexes

    def get_closest_lane_index(self, position, return_all=False):
        return self._graph_helper.get(position, return_all)

    def bfs_paths(self, start: str, goal: str) -> List[List[str]]:
        """
        Breadth-first search of all routes from start to goal.

        :param start: starting node
        :param goal: goal node
        :return: list of paths from start to goal.
        """
        queue = [(start, [start])]
        while queue:
            (node, path) = queue.pop(0)
            if node not in self.graph:
                yield []
            for _next in set(self.graph[node].keys()) - set(path):
                if _next == goal:
                    yield path + [_next]
                elif _next in self.graph:
                    queue.append((_next, path + [_next]))

    def shortest_path(self, start: str, goal: str) -> List[str]:
        """
        Breadth-first search of shortest checkpoints from start to goal.

        :param start: starting node
        :param goal: goal node
        :return: shortest checkpoints from start to goal.
        """
        start_road_node = start[0]
        assert start != goal
        return next(self.bfs_paths(start_road_node, goal), [])

    def get_map_features(self, interval=2):
        from metadrive.type import MetaDriveType

        ret = {}
        for _from, _to_dict in self.graph.items():
            for _to, lanes in _to_dict.items():
                for k, lane in enumerate(lanes):
                    ret["{}".format(lane.index)] = {
                        SD.POLYLINE: lane.get_polyline(interval),
                        SD.POLYGON: lane.polygon,
                        SD.TYPE: MetaDriveType.LANE_SURFACE_STREET,
                        "speed_limit_kmh": lane.speed_limit
                    }
        return ret

    def get_all_lanes(self):
        """
        This function will return all lanes in the road network
        :return: list of lanes
        """
        ret = []
        for _from, _to_dict in self.graph.items():
            for _to, lanes in _to_dict.items():
                ret += lanes
        return ret



import copy
import os

import numpy as np
import pytest

from metadrive import MetaDriveEnv
from metadrive.constants import TerminationState

# Key: scenario name, value: environmental config
blackbox_test_configs = dict(
    default=dict(),
    random_traffic=dict(random_traffic=True),
    large_seed=dict(start_seed=1000000),
    traffic_density_0=dict(traffic_density=0),
    traffic_density_1=dict(traffic_density=1),
    decision_repeat_50=dict(decision_repeat=50),
    map_7=dict(map=7),
    map_30=dict(map=30),
    map_CCC=dict(map="CCC"),
    envs_100=dict(num_scenarios=100),
    envs_1000=dict(num_scenarios=1000),
    envs_10000=dict(num_scenarios=10000),
    envs_100000=dict(num_scenarios=100000),
    no_lidar0={"vehicle_config": dict(lidar=dict(num_lasers=0, distance=0, num_others=0))},
    no_lidar1={"vehicle_config": dict(lidar=dict(num_lasers=0, distance=10, num_others=0))},
    no_lidar2={"vehicle_config": dict(lidar=dict(num_lasers=10, distance=0, num_others=0))},
    no_lidar3={"vehicle_config": dict(lidar=dict(num_lasers=0, distance=0, num_others=10))},
    no_lidar4={"vehicle_config": dict(lidar=dict(num_lasers=10, distance=10, num_others=0))},
    no_lidar5={"vehicle_config": dict(lidar=dict(num_lasers=10, distance=0, num_others=10))},
    no_lidar6={"vehicle_config": dict(lidar=dict(num_lasers=0, distance=10, num_others=10))},
    no_lidar7={"vehicle_config": dict(lidar=dict(num_lasers=10, distance=10, num_others=10))},
)

pid_control_config = dict(num_scenarios=1, start_seed=5, map="CrXROSTR", traffic_density=0.0, use_render=False)

info_keys = [
    "cost", "velocity", "steering", "acceleration", "step_reward", TerminationState.CRASH_VEHICLE,
    TerminationState.OUT_OF_ROAD, TerminationState.SUCCESS
]

assert "__init__.py" not in os.listdir(os.path.dirname(__file__)), "Please remove __init__.py in tests directory."


def _act(env, action):
    assert env.action_space.contains(action)
    obs, reward, terminated, truncated, info = env.step(action)
    assert env.observation_space.contains(obs)
    assert np.isscalar(reward)
    assert isinstance(info, dict)
    for k in info_keys:
        assert k in info


@pytest.mark.parametrize("config", list(blackbox_test_configs.values()), ids=list(blackbox_test_configs.keys()))
def test_pgdrive_env_blackbox(config):
    env = MetaDriveEnv(config=copy.deepcopy(config))
    try:
        obs, _ = env.reset()
        assert env.observation_space.contains(obs)
        _act(env, env.action_space.sample())
        for x in [-1, 0, 1]:
            env.reset()
            for y in [-1, 0, 1]:
                _act(env, [x, y])
    finally:
        env.close()


if __name__ == '__main__':
    pytest.main(["-sv", "test_metadrive_env.py"])



"""
This script defines a custom environment with single block: inramp.
"""
import random

import numpy as np

from metadrive import MetaDriveEnv

if __name__ == "__main__":
    config = dict(
        use_render=True,
        manual_control=True,
        start_seed=random.randint(0, 1000),

        # Solution 1: use easy config to customize the map
        # map="r",  # seven block

        # Solution 2: you can define more complex map config
        map_config=dict(lane_num=1, type="block_sequence", config="r")
    )

    env = MetaDriveEnv(config)
    try:
        o, _ = env.reset()
        env.vehicle.expert_takeover = True
        assert isinstance(o, np.ndarray)
        print("The observation is an numpy array with shape: ", o.shape)
        for i in range(1, 1000000000):
            o, r, tm, tc, info = env.step([0, 0])
            if (tm or tc) and info["arrive_dest"]:
                env.reset()
                env.current_track_vehicle.expert_takeover = True
    except:
        pass
    finally:
        env.close()



import json

from metadrive import MetaDriveEnv
from metadrive.utils import recursive_equal
from metadrive.utils.pg.generate_maps import generate_maps


def _test_loaded_map_alignment():
    raise DeprecationWarning("We do not generate maps from file now!")
    # Generate the second round
    # for seed in [0, 1, 2, 100, 200, 300, 9999]:
    for seed in [0, 1, 2, 99]:
        env_config = {"start_seed": seed, "num_scenarios": 1}
        generate_maps(MetaDriveEnv, env_config.copy(), json_file_path="seed{}_v1.json".format(seed))
        # generate_maps(MetaDriveEnv, env_config.copy(), json_file_path="seed{}_v2.json".format(seed))
        with open("seed{}_v1.json".format(seed), "r") as f:
            saved_v1 = json.load(f)
        # with open("seed{}_v2.json".format(seed), "r") as f:
        #     saved_v2 = json.load(f)

        e = MetaDriveEnv(env_config.copy())
        e.reset()
        map_data_realtime_load = e.current_map.get_meta_data()
        map_data_in_json = saved_v1["map_data"][str(seed)]
        e.close()

        e = MetaDriveEnv({"start_seed": seed, "num_scenarios": 1})
        e.reset()
        map_data_realtime_generate = e.current_map.get_meta_data()
        e.close()

        e = MetaDriveEnv({"start_seed": seed, "num_scenarios": 10})
        e.reset(seed=seed)
        map_data_realtime_generate_in_multiple_maps = e.current_map.get_meta_data()
        e.close()

        # Assert v1 and v2 have same maps
        # recursive_equal(saved_v1, saved_v2, True)

        # Assert json and loaded maps are same
        recursive_equal(map_data_in_json, map_data_realtime_load, True)

        # Assert json and generated maps are same
        recursive_equal(map_data_in_json, map_data_realtime_generate, True)

        # Assert json and generated maps in
        recursive_equal(map_data_in_json, map_data_realtime_generate_in_multiple_maps, True)

    # print(saved_v1)


def test_map_buffering():
    env_config = {"num_scenarios": 5}
    e = MetaDriveEnv(env_config)
    try:
        for i in range(10):
            e.reset()
        assert any([v is not None for v in e.engine.map_manager.maps.values()])
    finally:
        e.close()


if __name__ == '__main__':
    # test_loaded_map_alignment()
    test_map_buffering()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import print_function
from rplibs.six.moves import range  # pylint: disable=import-error

import math

from rplibs.pyqt_imports import *  # noqa


class CurveWidget(QWidget):
    """ This is a resizeable Widget which shows an editable curve which can
    be modified. """
    def __init__(self, parent):
        """ Constructs the CurveWidget, we start with an initial curve """
        QWidget.__init__(self, parent)
        self.setFocusPolicy(Qt.ClickFocus)
        self._curves = []

        # Store current display time
        self._current_time = 0.5

        # Widget render constants
        self._cv_point_size = 3
        self._legend_border = 52
        self._bar_h = 30

        # Currently dragged control point, format is:
        # (CurveIndex, PointIndex, Drag-Offset (x,y))
        self._drag_point = None
        self._drag_time = 0.0

        # Currently selected control point, format is:
        # (CurveIndex, PointIndex)
        self._selected_point = None

        self._unit_processor = lambda v: str(round(v, 2))
        self._change_handler = lambda: None

    def set_unit_processor(self, proc):
        """ Sets the function which gets called to map values from 0 .. 1 to
        values like 10% to 30% """
        self._unit_processor = proc

    def set_change_handler(self, handler):
        """ Sets a function which gets called when the data in this widget got
        edited """
        self._change_handler = handler

    def paintEvent(self, e):  # noqa
        """ Internal QT paint event, draws the entire widget """
        qp = QPainter()
        qp.begin(self)
        self._draw(qp)
        qp.end()

    def mousePressEvent(self, QMouseEvent):  # noqa
        """ Internal mouse-press handler """
        self._drag_point = None
        self._selected_point = None
        mouse_pos = QMouseEvent.pos()
        mouse_x = mouse_pos.x() - self._legend_border
        mouse_y = mouse_pos.y() - self._bar_h

        for index, curve in enumerate(self._curves):

            # Check for clicks on control points
            for cv_index, (x, y) in enumerate(curve.control_points):
                point_x = self._get_x_value_for(x)
                point_y = self._get_y_value_for(y) - self._bar_h
                if abs(point_x - mouse_x) < self._cv_point_size + 6:
                    if (abs(point_y - mouse_y)) < self._cv_point_size + 6:
                        drag_x_offset = mouse_x - point_x
                        drag_y_offset = mouse_y - point_y
                        mpos_relx = float(mouse_x) / (self.width() - self._legend_border)
                        self._drag_point = (index, cv_index, (drag_x_offset, drag_y_offset))
                        self._drag_time = mpos_relx
                        self._selected_point = (index, cv_index)

            # If still no intersection, check if we clicked a curve
            if mouse_x > 0 and mouse_x < self.width() - self._legend_border:
                if mouse_y > 0 and mouse_y < self.height() - self._legend_border:
                    mpos_relx = float(mouse_x) / (self.width() - self._legend_border)
                    curve_py = curve.get_value(mpos_relx)
                    curve_offy = self._get_y_value_for(curve_py) - self._bar_h

                    if abs(curve_offy - mouse_y) < 8 and self._selected_point is None:
                        # Clicked on curve, spawn new point
                        cv_index = curve.append_cv(mpos_relx, curve_py)

                        self._selected_point = (index, cv_index)
                        self._drag_point = (index, cv_index, (0, 0))
                        self._drag_time = mpos_relx
                        self._change_handler()
        self.update()

    def mouseReleaseEvent(self, event):  # noqa
        """ Internal mouse-release handler """
        self._drag_point = None
        self._drag_time = -1
        self.update()

    def mouseMoveEvent(self, event):  # noqa
        """ Internal mouse-move handler """
        if len(self._curves) < 1:
            return

        if self._drag_point is not None:
            mouse_x = event.pos().x() - self._drag_point[2][0] - self._legend_border
            mouse_y = event.pos().y() - self._drag_point[2][1] - self._bar_h

            # Convert to local coordinate
            local_x = max(0, min(1, mouse_x / float(self.width() - self._legend_border)))
            local_y = mouse_y / float(self.height() - self._legend_border - self._bar_h)
            local_y = 1 - max(0, min(1, local_y))

            self._drag_time = local_x

            # Set new point data
            self._curves[self._drag_point[0]].set_cv_value(self._drag_point[1], local_x, local_y)

            # Redraw curve
            self._curves[self._drag_point[0]].build_curve()
            self.update()
            self._change_handler()

    def keyPressEvent(self, event):  # noqa
        """ Internal keypress handler """
        # Delete anchor point
        if event.key() == Qt.Key_Delete:
            self.delete_current_point()

    def delete_current_point(self):
        """ Deletes the currently selected point """
        if self._selected_point is not None:
            self._curves[self._selected_point[0]].remove_cv(self._selected_point[1])
            self._selected_point = None
            self._drag_point = None
            self.update()
            self._change_handler()

    def set_curves(self, curves):
        """ Sets the list of displayed curves """
        self._selected_point = None
        self._drag_point = None
        self._curves = curves
        self.update()

    def _get_y_value_for(self, local_value):
        """ Converts a value from 0 to 1 to a value from 0 .. canvas height """
        local_value = max(0, min(1.0, 1.0 - local_value))
        local_value *= self.height() - self._legend_border - self._bar_h
        local_value += self._bar_h
        return local_value

    def _get_x_value_for(self, local_value):
        """ Converts a value from 0 to 1 to a value from 0 .. canvas width """
        local_value = max(0, min(1.0, local_value))
        local_value *= self.width() - self._legend_border
        return local_value

    def set_current_time(self, local_time):
        """ Sets the current displayed time, should range from 0 to 1 """
        self._current_time = max(0.0, min(1.0, local_time))
        self.update()

    def _draw(self, painter):
        """ Internal method to draw the widget """

        canvas_width = self.width() - self._legend_border
        canvas_height = self.height() - self._legend_border - self._bar_h

        # Draw field background
        # painter.setPen(QColor(200, 200, 200))
        # painter.setBrush(QColor(230, 230, 230))
        # painter.drawRect(0, 0, self.width() - 1, self.height() - 1)

        # Draw legend

        # Compute amount of horizontal / vertical lines
        num_vert_lines = 12  # 24 / 12 = 2, one entry per 2 hours
        line_spacing_x = (self.width() - self._legend_border) / num_vert_lines
        line_spacing_y = (self.height() - self._legend_border - self._bar_h) / 20.0
        num_horiz_lines = int(math.ceil(canvas_height / float(line_spacing_y)) + 1)

        # Draw vertical lines
        painter.setPen(QColor(200, 200, 200))
        for i in range(num_vert_lines + 1):
            line_pos = i * line_spacing_x + self._legend_border - 1
            painter.drawLine(line_pos, self._bar_h, line_pos, canvas_height + self._bar_h)

        # Draw horizontal lines
        painter.setPen(QColor(200, 200, 200))
        for i in range(num_horiz_lines):
            line_pos = canvas_height - i * line_spacing_y + self._bar_h
            painter.drawLine(self._legend_border, line_pos, self.width(), line_pos)

        # Draw vetical legend labels
        painter.setPen(QColor(120, 120, 120))
        for i in range(num_horiz_lines):
            line_pos = canvas_height - i * line_spacing_y + self._bar_h
            # painter.drawText(6, line_pos + 3, str(round(float(i) / (num_horiz_lines-1), 2)))
            painter.drawText(6, line_pos + 3, self._unit_processor(float(i) / (num_horiz_lines - 1)))

        # Draw horizontal legend labels
        for i in range(num_vert_lines + 1):
            line_pos = i * line_spacing_x + self._legend_border
            offpos_x = -14
            if i == 0:
                offpos_x = -2
            elif i == num_vert_lines:
                offpos_x = -27
            time_string = str(int(float(i) / num_vert_lines * 24)).zfill(2) + ":00"
            painter.drawText(line_pos + offpos_x, canvas_height + self._bar_h + 18, time_string)

        # Draw curve
        for index, curve in enumerate(self._curves):
            painter.setPen(QColor(*curve.color))
            last_value = 0
            for i in range(canvas_width):
                rel_offset = i / (canvas_width - 1.0)
                curve_height = self._get_y_value_for(curve.get_value(rel_offset))

                if i == 0:
                    last_value = curve_height

                painter.drawLine(self._legend_border + i - 1, last_value, self._legend_border + i, curve_height)
                last_value = curve_height

            # Draw the CV points of the curve
            painter.setBrush(QColor(240, 240, 240))

            for cv_index, (x, y) in enumerate(curve.control_points):
                offs_x = x * canvas_width + self._legend_border
                offs_y = (1 - y) * canvas_height + self._bar_h

                if (self._selected_point and self._selected_point[0] == index and self._selected_point[1] == cv_index):
                    painter.setPen(QColor(255, 0, 0))
                else:
                    painter.setPen(QColor(100, 100, 100))
                painter.drawRect(
                    offs_x - self._cv_point_size, offs_y - self._cv_point_size, 2 * self._cv_point_size,
                    2 * self._cv_point_size
                )

        # Draw bar background
        bar_half_height = 4
        bar_top_pos = 10

        painter.setBrush(QColor(255, 0, 0))
        painter.setPen(QColor(110, 110, 110))

        painter.drawRect(
            self._legend_border - 1, bar_top_pos - 1,
            self.width() - self._legend_border, 2 * bar_half_height + 2
        )

        # Draw bar
        if len(self._curves) == 0:
            return

        if len(self._curves) == 1:
            bar_curve = self._curves[0]
        else:
            bar_curve = self._curves[0:3]

        for i in range(canvas_width - 1):
            xpos = self._legend_border + i
            relv = float(i) / float(canvas_width)

            if len(self._curves) == 1:
                val = max(0, min(255, int(bar_curve.get_value(relv) * 255.0)))
                painter.setPen(QColor(val, val, val))
            else:
                r = max(0, min(255, int(bar_curve[0].get_value(relv) * 255.0)))
                g = max(0, min(255, int(bar_curve[1].get_value(relv) * 255.0)))
                b = max(0, min(255, int(bar_curve[2].get_value(relv) * 255.0)))
                painter.setPen(QColor(r, g, b))
            painter.drawLine(xpos, bar_top_pos, xpos, bar_top_pos + 2 * bar_half_height)

        # Draw selected time
        if self._drag_point:
            painter.setBrush(QColor(200, 200, 200))
            painter.setPen(QColor(90, 90, 90))
            offs_x = max(0, min(canvas_width + 10, self._drag_time * canvas_width + self._legend_border - 19))
            offs_y = self.height() - self._legend_border
            minutes = int(self._drag_time * 24 * 60)
            hours = minutes / 60
            minutes = minutes % 60
            painter.drawRect(offs_x, self.height() - self._legend_border + 5, 40, 20)
            painter.drawText(offs_x + 7, offs_y + 20, "{:02}:{:02}".format(hours, minutes))

            painter.setPen(QColor(150, 150, 150))
            painter.drawLine(offs_x + 19, bar_top_pos + 15, offs_x + 19, self.height() - self._legend_border + 5)

        # Display current time
        pen = QPen()
        pen.setColor(QColor(255, 100, 100))
        pen.setStyle(Qt.DashLine)
        painter.setPen(pen)

        xoffs = self._legend_border + self._current_time * (canvas_width - 1)
        painter.drawLine(xoffs, self._bar_h, xoffs, self._bar_h + canvas_height)

        # Draw usage hints
        painter.setPen(QColor(100, 100, 100))
        painter.drawText(
            5,
            self.height() - 2, "Click on the curve to add new control points, click and drag "
            "existing points to move them."
        )



# DiamondSquare.py
# Written by Hailey K Buckingham
# On github, buckinha/DiamondSquare
#

import random
import math
import numpy as np


def diamond_square(
    shape: (int, int),
    min_height: [float or int],
    max_height: [float or int],
    roughness: [float or int],
    random_seed=None,
    uint16=True,
    minus_mean=True
):
    """Runs a diamond square algorithm and returns an array (or list) with the landscape

        An important difference (possibly) between this, and other implementations of the
    diamond square algorithm is how I use the roughness parameter. For each "perturbation"
    I pull a random number from a uniform distribution between min_height and max_height.
    I then take the weighted average between that value, and the average value of the
    "neighbors", whether those be in the diamond or in the square step, as normal. The
    weights used for the weighted sum are (roughness) and (1-roughness) for the random
    number and the average, respectively, where roughness is a float that always falls
    between 0 and 1.
        The roughness value used in each iteration is based on the roughness parameter
    passed in, and is computed as follows:

        this_iteration_roughness = roughness**iteration_number

    where the first iteration has iteration_number = 0. The first roughness value
    actually used (in the very first diamond and square step) is roughness**0 = 1. Thus,
    the values for those first diamond and square step entries will be entirely random.
    This effectively means that I am seeding with A 3x3 grid of random values, rather
    than with just the four corners.
        As the process continues, the weight placed on the random number draw falls from
    the original value of 1, to roughness**1, to roughness**2, and so on, ultimately
    approaching 0. This means that the values of new cells will slowly shift from being
    purely random, to pure averages.


    OTHER NOTES:
    Internally, all heights are between 0 and 1, and are rescaled at the end.


    PARAMETERS
    ----------
    :param shape
        tuple of ints, (int, int): the shape of the resulting landscape

    :param min_height
        Int or Float: The minimum height allowed on the landscape

    :param max_height
        Int or Float: The maximum height allowed on the landscape

    :param roughness
        Float with value between 0 and 1, reflecting how bumpy the landscape should be.
        Values near 1 will result in landscapes that are extremely rough, and have almost no
        cell-to-cell smoothness. Values near zero will result in landscapes that are almost
        perfectly smooth.

        Values above 1.0 will be interpreted as 1.0
        Values below 0.0 will be interpreted as 0.0

    :param random_seed
        Any value. Defaults to None. If a value is given, the algorithm will use it to seed the random
        number generator, ensuring replicability.

    :param as_ndarray
        Bool: whether the landscape should be returned as a numpy array. If set
        to False, the method will return list of lists.

    :returns [list] or nd_array
    """

    # sanitize inputs
    if roughness > 1:
        roughness = 1.0
    if roughness < 0:
        roughness = 0.0

    working_shape, iterations = _get_working_shape_and_iterations(shape)

    # create the array
    diamond_square_array = np.full(working_shape, -1, dtype='float')

    # seed the random number generator
    random.seed(random_seed)

    # seed the corners
    diamond_square_array[0, 0] = random.uniform(0, 1)
    diamond_square_array[working_shape[0] - 1, 0] = random.uniform(0, 1)
    diamond_square_array[0, working_shape[1] - 1] = random.uniform(0, 1)
    diamond_square_array[working_shape[0] - 1, working_shape[1] - 1] = random.uniform(0, 1)

    # do the algorithm
    for i in range(iterations):
        r = math.pow(roughness, i)

        step_size = math.floor((working_shape[0] - 1) / math.pow(2, i))

        _diamond_step(diamond_square_array, step_size, r)
        _square_step(diamond_square_array, step_size, r)

    # rescale the array to fit the min and max heights specified
    diamond_square_array = min_height + (diamond_square_array * (max_height - min_height))

    # trim array, if needed
    final_array = diamond_square_array[:shape[0], :shape[1]]

    if minus_mean:
        final_array -= np.mean(final_array)

    if uint16:
        final_array = final_array.astype(np.uint16)

    return final_array


def _get_working_shape_and_iterations(requested_shape, max_power_of_two=13):
    """Returns the necessary size for a square grid which is usable in a DS algorithm.

    The Diamond Square algorithm requires a grid of size n x n where n = 2**x + 1, for any
    integer value of x greater than two. To accomodate a requested map size other than these
    dimensions, we simply create the next largest n x n grid which can entirely contain the
    requested size, and return a subsection of it.

    This method computes that size.

    PARAMETERS
    ----------
    requested_shape
        A 2D list-like object reflecting the size of grid that is ultimately desired.

    max_power_of_two
        an integer greater than 2, reflecting the maximum size grid that the algorithm can EVER
        attempt to make, even if the requested size is too big. This limits the algorithm to
        sizes that are manageable, unless the user really REALLY wants to have a bigger one.
        The maximum grid size will have an edge of size  (2**max_power_of_two + 1)

    RETURNS
    -------
    An integer of value n, as described above.
    """
    if max_power_of_two < 3:
        max_power_of_two = 3

    largest_edge = max(requested_shape)

    for power in range(1, max_power_of_two + 1):
        d = (2**power) + 1
        if largest_edge <= d:
            return (d, d), power

    # failsafe: no values in the dimensions array were allowed, so print a warning and return
    # the maximum size.
    d = 2**max_power_of_two + 1
    print("DiamondSquare Warning: Requested size was too large. Grid of size {0} returned" "".format(d))
    return (d, d), max_power_of_two


def _diamond_step(DS_array, step_size, roughness):
    """Does the diamond step for a given iteration.

    During the diamond step, the diagonally adjacent cells are filled:

    Value   None   Value   None   Value  ...

    None   FILLING  None  FILLING  None  ...

    Value   None   Value   None   Value  ...

    ...     ...     ...     ...    ...   ...

    So we'll step with increment step_size over BOTH axes

    """
    # calculate where all the diamond corners are (the ones we'll be filling)
    half_step = math.floor(step_size / 2)
    x_steps = range(half_step, DS_array.shape[0], step_size)
    y_steps = x_steps[:]

    for i in x_steps:
        for j in y_steps:
            if DS_array[i, j] == -1.0:
                DS_array[i, j] = _diamond_displace(DS_array, i, j, half_step, roughness)


def _square_step(DS_array, step_size, roughness):
    """Does the square step for a given iteration.

    During the diamond step, the diagonally adjacent cells are filled:

     Value    FILLING    Value    FILLING   Value   ...

    FILLING   DIAMOND   FILLING   DIAMOND  FILLING  ...

     Value    FILLING    Value    FILLING   Value   ...

      ...       ...       ...       ...      ...    ...

    So we'll step with increment step_size over BOTH axes

    """

    # doing this in two steps: the first, where the every other column is skipped
    # and the second, where every other row is skipped. For each, iterations along
    # the half-steps go vertically or horizontally, respectively.

    # set the half-step for the calls to square_displace
    half_step = math.floor(step_size / 2)

    # vertical step
    steps_x_vert = range(half_step, DS_array.shape[0], step_size)
    steps_y_vert = range(0, DS_array.shape[1], step_size)

    # horizontal step
    steps_x_horiz = range(0, DS_array.shape[0], step_size)
    steps_y_horiz = range(half_step, DS_array.shape[1], step_size)

    for i in steps_x_horiz:
        for j in steps_y_horiz:
            DS_array[i, j] = _square_displace(DS_array, i, j, half_step, roughness)

    for i in steps_x_vert:
        for j in steps_y_vert:
            DS_array[i, j] = _square_displace(DS_array, i, j, half_step, roughness)


def _diamond_displace(DS_array, i, j, half_step, roughness):
    """
    defines the midpoint displacement for the diamond step
    :param DS_array:
    :param i:
    :param j:
    :param half_step:
    :param roughness:
    :return:
    """
    ul = DS_array[i - half_step, j - half_step]
    ur = DS_array[i - half_step, j + half_step]
    ll = DS_array[i + half_step, j - half_step]
    lr = DS_array[i + half_step, j + half_step]

    ave = (ul + ur + ll + lr) / 4.0

    rand_val = random.uniform(0, 1)

    return (roughness * rand_val) + (1.0 - roughness) * ave


def _square_displace(DS_array, i, j, half_step, roughness):
    """
    Defines the midpoint displacement for the square step

    :param DS_array:
    :param i:
    :param j:
    :param half_step:
    :param roughness:
    :return:
    """
    _sum = 0.0
    divide_by = 4

    # check cell "above"
    if i - half_step >= 0:
        _sum += DS_array[i - half_step, j]
    else:
        divide_by -= 1

    # check cell "below"
    if i + half_step < DS_array.shape[0]:
        _sum += DS_array[i + half_step, j]
    else:
        divide_by -= 1

    # check cell "left"
    if j - half_step >= 0:
        _sum += DS_array[i, j - half_step]
    else:
        divide_by -= 1

    # check cell "right"
    if j + half_step < DS_array.shape[0]:
        _sum += DS_array[i, j + half_step]
    else:
        divide_by -= 1

    ave = _sum / divide_by

    rand_val = random.uniform(0, 1)

    return (roughness * rand_val) + (1.0 - roughness) * ave



"""
This script demonstrates how to train a set of policies under different number of training scenarios and test them
in the same test set using rllib.

We verified this script with ray==2.2.0. Please report to use if you find newer version of ray is not compatible with
this script.
"""
import argparse
import copy
import logging
from typing import Dict

import numpy as np

from metadrive import MetaDriveEnv
from metadrive.envs.gym_wrapper import createGymWrapper

try:
    import ray
    from ray import tune

    from ray.tune import CLIReporter
    from ray.rllib.algorithms.callbacks import DefaultCallbacks
    from ray.rllib.env import BaseEnv
    from ray.rllib.evaluation import MultiAgentEpisode, RolloutWorker
    from ray.rllib.policy import Policy
except ImportError:
    ray = None
    raise ValueError("Please install ray through 'pip install ray'.")


class DrivingCallbacks(DefaultCallbacks):
    def on_episode_start(
        self, *, worker: RolloutWorker, base_env: BaseEnv, policies: Dict[str, Policy], episode: MultiAgentEpisode,
        env_index: int, **kwargs
    ):
        episode.user_data["velocity"] = []
        episode.user_data["steering"] = []
        episode.user_data["step_reward"] = []
        episode.user_data["acceleration"] = []
        episode.user_data["cost"] = []

    def on_episode_step(
        self, *, worker: RolloutWorker, base_env: BaseEnv, episode: MultiAgentEpisode, env_index: int, **kwargs
    ):
        info = episode.last_info_for()
        if info is not None:
            episode.user_data["velocity"].append(info["velocity"])
            episode.user_data["steering"].append(info["steering"])
            episode.user_data["step_reward"].append(info["step_reward"])
            episode.user_data["acceleration"].append(info["acceleration"])
            episode.user_data["cost"].append(info["cost"])

    def on_episode_end(
        self, worker: RolloutWorker, base_env: BaseEnv, policies: Dict[str, Policy], episode: MultiAgentEpisode,
        **kwargs
    ):
        arrive_dest = episode.last_info_for()["arrive_dest"]
        crash = episode.last_info_for()["crash"]
        out_of_road = episode.last_info_for()["out_of_road"]
        max_step_rate = not (arrive_dest or crash or out_of_road)
        episode.custom_metrics["success_rate"] = float(arrive_dest)
        episode.custom_metrics["crash_rate"] = float(crash)
        episode.custom_metrics["out_of_road_rate"] = float(out_of_road)
        episode.custom_metrics["max_step_rate"] = float(max_step_rate)
        episode.custom_metrics["velocity_max"] = float(np.max(episode.user_data["velocity"]))
        episode.custom_metrics["velocity_mean"] = float(np.mean(episode.user_data["velocity"]))
        episode.custom_metrics["velocity_min"] = float(np.min(episode.user_data["velocity"]))
        episode.custom_metrics["steering_max"] = float(np.max(episode.user_data["steering"]))
        episode.custom_metrics["steering_mean"] = float(np.mean(episode.user_data["steering"]))
        episode.custom_metrics["steering_min"] = float(np.min(episode.user_data["steering"]))
        episode.custom_metrics["acceleration_min"] = float(np.min(episode.user_data["acceleration"]))
        episode.custom_metrics["acceleration_mean"] = float(np.mean(episode.user_data["acceleration"]))
        episode.custom_metrics["acceleration_max"] = float(np.max(episode.user_data["acceleration"]))
        episode.custom_metrics["step_reward_max"] = float(np.max(episode.user_data["step_reward"]))
        episode.custom_metrics["step_reward_mean"] = float(np.mean(episode.user_data["step_reward"]))
        episode.custom_metrics["step_reward_min"] = float(np.min(episode.user_data["step_reward"]))
        episode.custom_metrics["cost"] = float(sum(episode.user_data["cost"]))

    def on_train_result(self, *, algorithm, result: dict, **kwargs):
        result["success"] = np.nan
        result["crash"] = np.nan
        result["out"] = np.nan
        result["max_step"] = np.nan
        result["length"] = result["episode_len_mean"]
        result["cost"] = np.nan
        if "custom_metrics" not in result:
            return

        if "success_rate_mean" in result["custom_metrics"]:
            result["success"] = result["custom_metrics"]["success_rate_mean"]
            result["crash"] = result["custom_metrics"]["crash_rate_mean"]
            result["out"] = result["custom_metrics"]["out_of_road_rate_mean"]
            result["max_step"] = result["custom_metrics"]["max_step_rate_mean"]
        if "cost_mean" in result["custom_metrics"]:
            result["cost"] = result["custom_metrics"]["cost_mean"]


def train(
    trainer,
    config,
    stop,
    exp_name,
    num_gpus=0,
    test_mode=False,
    checkpoint_freq=10,
    keep_checkpoints_num=None,
    custom_callback=None,
    max_failures=5,
    **kwargs
):
    ray.init(
        num_gpus=num_gpus,
        logging_level=logging.ERROR if not test_mode else logging.DEBUG,
        log_to_driver=test_mode,
    )
    used_config = {
        "callbacks": custom_callback if custom_callback else DrivingCallbacks,  # Must Have!
        "log_level": "DEBUG" if test_mode else "WARN",
    }
    used_config.update(config)
    config = copy.deepcopy(used_config)

    if not isinstance(stop, dict) and stop is not None:
        assert np.isscalar(stop)
        stop = {"timesteps_total": int(stop)}

    if keep_checkpoints_num is not None and not test_mode:
        assert isinstance(keep_checkpoints_num, int)
        kwargs["keep_checkpoints_num"] = keep_checkpoints_num
        kwargs["checkpoint_score_attr"] = "episode_reward_mean"

    metric_columns = CLIReporter.DEFAULT_COLUMNS.copy()
    progress_reporter = CLIReporter(metric_columns=metric_columns)
    progress_reporter.add_metric_column("success")
    progress_reporter.add_metric_column("crash")
    progress_reporter.add_metric_column("out")
    progress_reporter.add_metric_column("max_step")
    progress_reporter.add_metric_column("length")
    progress_reporter.add_metric_column("cost")
    kwargs["progress_reporter"] = progress_reporter

    if "verbose" not in kwargs:
        kwargs["verbose"] = 1 if not test_mode else 2

    # start training
    analysis = tune.run(
        trainer,
        name=exp_name,
        checkpoint_freq=checkpoint_freq,
        checkpoint_at_end=True if "checkpoint_at_end" not in kwargs else kwargs.pop("checkpoint_at_end"),
        stop=stop,
        config=config,
        max_failures=max_failures if not test_mode else 0,
        reuse_actors=False,
        local_dir=".",
        **kwargs
    )
    return analysis


def get_train_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--exp-name", type=str, default="generalization_experiment")
    parser.add_argument("--num-gpus", type=int, default=0)
    return parser


if __name__ == '__main__':
    args = get_train_parser().parse_args()
    exp_name = args.exp_name
    stop = int(1000_0000)
    config = dict(

        # ===== Training Environment =====
        # Train the policies in scenario sets with different number of scenarios.
        env=createGymWrapper(MetaDriveEnv),
        env_config=dict(
            num_scenarios=tune.grid_search([1, 3, 5, 1000]),
            start_seed=tune.grid_search([5000]),
            random_traffic=False,
            traffic_density=tune.grid_search([0.1, 0.3])
        ),

        # ===== Framework =====
        framework="torch",

        # ===== Evaluation =====
        # Evaluate the trained policies in unseen 200 scenarios.
        evaluation_interval=2,
        evaluation_num_episodes=40,
        metrics_smoothing_episodes=200,
        evaluation_config=dict(env_config=dict(num_scenarios=200, start_seed=0)),
        evaluation_num_workers=5,

        # ===== Training =====
        # Hyper-parameters for PPO
        horizon=1000,
        rollout_fragment_length=200,
        sgd_minibatch_size=256,
        train_batch_size=20000,
        num_sgd_iter=10,
        lr=3e-4,
        num_workers=5,
        **{"lambda": 0.95},

        # ===== Resources Specification =====
        num_gpus=0.25 if args.num_gpus != 0 else 0,
        num_cpus_per_worker=0.2,
        num_cpus_for_driver=0.5,
    )

    train(
        "PPO",
        exp_name=exp_name,
        keep_checkpoints_num=5,
        stop=stop,
        config=config,
        num_gpus=args.num_gpus,
        test_mode=False
    )



from metadrive.utils.config import Config


def test_recursive_config():
    c = Config({"aa": {"bb": {"cc": 100}}})
    assert c.aa.bb.cc == 100
    assert isinstance(c.aa, Config)
    assert isinstance(c.aa.bb, Config)
    assert isinstance(c.aa.bb.cc, int)

    c.update({"aa": {"bb": {"cc": 101}}})
    assert c.aa.bb.cc == 101

    try:
        c.update({"aa": {"bb": 102}}, allow_add_new_key=False)
    except TypeError:
        pass
    else:
        raise ValueError()

    try:
        c.update({"aa": {"bbd": 102}}, allow_add_new_key=False)
    except KeyError:
        pass
    else:
        raise ValueError()

    c.update({"aa": {"bbd": 103}})
    assert c.aa.bbd == 103
    assert c.aa.bb.cc == 101

    c.update({"aa": {"bb": 102}})
    assert c.aa.bb == 102


def test_partially_update():
    c = Config({"aa": {"bb": {"cc": 100}}})

    try:
        c.update({"aa": {"bb": {"dd": 101}}}, allow_add_new_key=False)
    except KeyError:
        pass
    else:
        raise ValueError()

    c.update({"aa": {"bb": {"dd": 101}}})
    assert c.aa.bb.cc == 100
    assert c.aa.bb.dd == 101


def test_config_identical():
    c = Config({"aa": {"bb": {"cc": 100}}})
    d = Config({"aa": {"bb": {"cc": 100}}})
    assert c.is_identical(c)
    assert c.is_identical(d)
    assert d.is_identical(c)
    assert d.is_identical(d)

    c.aa.bb.cc = 101
    assert c.is_identical(c)
    assert d.is_identical(d)
    assert not c.is_identical(d)
    assert not d.is_identical(c)

    c.update({"aa": {"bb": 10001}})
    assert c.is_identical(c)
    assert d.is_identical(d)
    assert not c.is_identical(d)
    assert not d.is_identical(c)


def test_dict_as_attribute():
    c = Config({"aa": {"bb": {"cc": 100, "ee": 102}}})
    c.update({"ff": {"gg": 104}})
    assert hasattr(c, "ff")


if __name__ == '__main__':
    # test_recursive_config()
    # test_partially_update()
    # test_config_identical()
    test_dict_as_attribute()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division

import sys
import math
import time

from panda3d.core import LVecBase2i, TransformState, RenderState, load_prc_file
from panda3d.core import PandaSystem, MaterialAttrib, WindowProperties
from panda3d.core import GeomTristrips, Vec4

from direct.showbase.ShowBase import ShowBase
from direct.stdpy.file import isfile

from metadrive.render_pipeline.rplibs.yaml import load_yaml_file_flat
from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.effect import Effect
from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.common_resources import CommonResources
from metadrive.render_pipeline.rpcore.native import TagStateManager, PointLight, SpotLight
from metadrive.render_pipeline.rpcore.render_target import RenderTarget
from metadrive.render_pipeline.rpcore.pluginbase.manager import PluginManager
from metadrive.render_pipeline.rpcore.pluginbase.day_manager import DayTimeManager

from metadrive.render_pipeline.rpcore.util.task_scheduler import TaskScheduler
from metadrive.render_pipeline.rpcore.util.network_communication import NetworkCommunication
from metadrive.render_pipeline.rpcore.util.ies_profile_loader import IESProfileLoader

from metadrive.render_pipeline.rpcore.gui.debugger import Debugger
from metadrive.render_pipeline.rpcore.gui.loading_screen import LoadingScreen

from metadrive.render_pipeline.rpcore.mount_manager import MountManager
from metadrive.render_pipeline.rpcore.stage_manager import StageManager
from metadrive.render_pipeline.rpcore.light_manager import LightManager

from metadrive.render_pipeline.rpcore.stages.ambient_stage import AmbientStage
from metadrive.render_pipeline.rpcore.stages.gbuffer_stage import GBufferStage
from metadrive.render_pipeline.rpcore.stages.final_stage import FinalStage
from metadrive.render_pipeline.rpcore.stages.downscale_z_stage import DownscaleZStage
from metadrive.render_pipeline.rpcore.stages.combine_velocity_stage import CombineVelocityStage
from metadrive.render_pipeline.rpcore.stages.upscale_stage import UpscaleStage


class RenderPipeline(RPObject):
    """ This is the main render pipeline class, it combines all components of
    the pipeline to form a working system. It does not do much work itself, but
    instead setups all the managers and systems to be able to do their work. """
    def __init__(self):
        """ Creates a new pipeline with a given showbase instance. This should
        be done before intializing the ShowBase, the pipeline will take care of
        that. If the showbase has been initialized before, have a look at
        the alternative initialization of the render pipeline (the first sample)."""
        RPObject.__init__(self)
        self._analyze_system()
        self.mount_mgr = MountManager(self)
        self.settings = {}
        self._applied_effects = []
        self._pre_showbase_initialized = False
        self._first_frame = None
        self.set_loading_screen_image("/$$rp/data/gui/loading_screen_bg.txo")

    def load_settings(self, path):
        """ Loads the pipeline configuration from a given filename. Usually
        this is the 'config/pipeline.ini' file. If you call this more than once,
        only the settings of the last file will be used. """
        self.settings = load_yaml_file_flat(path)

    def reload_shaders(self):
        """ Reloads all shaders. This will reload the shaders of all plugins,
        as well as the pipelines internally used shaders. Because of the
        complexity of some shaders, this operation take might take several
        seconds. Also notice that all applied effects will be lost, and instead
        the default effect will be set on all elements again. Due to this fact,
        this method is primarly useful for fast iterations when developing new
        shaders. """
        if self.settings["pipeline.display_debugger"]:
            self.debug("Reloading shaders ..")
            self.debugger.error_msg_handler.clear_messages()
            self.debugger.set_reload_hint_visible(True)
            self._showbase.graphicsEngine.render_frame()
            self._showbase.graphicsEngine.render_frame()
        self.tag_mgr.cleanup_states()
        self.stage_mgr.reload_shaders()
        self.light_mgr.reload_shaders()
        self._set_default_effect()
        self.plugin_mgr.trigger_hook("shader_reload")
        if self.settings["pipeline.display_debugger"]:
            self.debugger.set_reload_hint_visible(False)
        self._apply_custom_shaders()

    def _apply_custom_shaders(self):
        """ Re-applies all custom shaders the user applied, to avoid them getting
        removed when the shaders are reloaded """
        self.debug("Re-applying", len(self._applied_effects), "custom shaders")
        for args in self._applied_effects:
            self._internal_set_effect(*args)

    def pre_showbase_init(self):
        """ Setups all required pipeline settings and configuration which have
        to be set before the showbase is setup. This is called by create(),
        in case the showbase was not initialized, however you can (and have to)
        call it manually before you init your custom showbase instance.
        See the 00-Loading the pipeline sample for more information. """
        if not self.mount_mgr.is_mounted:
            self.debug("Mount manager was not mounted, mounting now ...")
            self.mount_mgr.mount()

        if not self.settings:
            self.debug("No settings loaded, loading from default location")
            self.load_settings("/$$rpconfig/pipeline.yaml")

        if not isfile("/$$rp/data/install.flag"):
            self.fatal("You didn't setup the pipeline yet! Please run setup.py.")

        load_prc_file("/$$rpconfig/panda3d-config.prc")
        self._pre_showbase_initialized = True

    def create(self, base=None):
        """ This creates the pipeline, and setups all buffers. It also
        constructs the showbase. The settings should have been loaded before
        calling this, and also the base and write path should have been
        initialized properly (see MountManager).

        If base is None, the showbase used in the RenderPipeline constructor
        will be used and initialized. Otherwise it is assumed that base is an
        initialized ShowBase object. In this case, you should call
        pre_showbase_init() before initializing the ShowBase"""

        start_time = time.time()
        self._init_showbase(base)

        if not self._showbase.win.gsg.supports_compute_shaders:
            self.fatal(
                "Sorry, your GPU does not support compute shaders! Make sure\n"
                "you have the latest drivers. If you already have, your gpu might\n"
                "be too old, or you might be using the open source drivers on linux."
            )

        self._init_globals()
        self.loading_screen.create()
        self._adjust_camera_settings()
        self._create_managers()
        self.plugin_mgr.load()
        self.daytime_mgr.load_settings()
        self.common_resources.write_config()
        self._init_debugger()

        self.plugin_mgr.trigger_hook("stage_setup")
        self.plugin_mgr.trigger_hook("post_stage_setup")

        self._create_common_defines()
        self._initialize_managers()
        self._create_default_skybox()

        self.plugin_mgr.trigger_hook("pipeline_created")

        self._listener = NetworkCommunication(self)
        self._set_default_effect()

        # Measure how long it took to initialize everything, and also store
        # when we finished, so we can measure how long it took to render the
        # first frame (where the shaders are actually compiled)
        init_duration = (time.time() - start_time)
        self._first_frame = time.process_time()
        self.debug(
            "Finished initialization in {:3.3f} s, first frame: {}".format(
                init_duration, Globals.clock.get_frame_count()
            )
        )

    def set_loading_screen_image(self, image_source):
        """ Tells the pipeline to use the default loading screen, which consists
        of a simple loading image. The image source should be a fullscreen
        16:9 image, and not too small, to avoid being blurred out. """
        self.loading_screen = LoadingScreen(self, image_source)

    def add_light(self, light):
        """ Adds a new light to the rendered lights, check out the LightManager
        add_light documentation for further information. """
        self.light_mgr.add_light(light)

    def remove_light(self, light):
        """ Removes a previously attached light, check out the LightManager
        remove_light documentation for further information. """
        self.light_mgr.remove_light(light)

    def load_ies_profile(self, filename):
        """ Loads an IES profile from a given filename and returns a handle which
        can be used to set an ies profile on a light """
        return self.ies_loader.load(filename)

    def _internal_set_effect(self, nodepath, effect_src, options=None, sort=30):
        """ Sets an effect to the given object, using the specified options.
        Check out the effect documentation for more information about possible
        options and configurations. The object should be a nodepath, and the
        effect will be applied to that nodepath and all nodepaths below whose
        current effect sort is less than the new effect sort (passed by the
        sort parameter). """
        effect = Effect.load(effect_src, options)
        if effect is None:
            return self.error("Could not apply effect")

        for i, stage in enumerate(("gbuffer", "shadow", "voxelize", "envmap", "forward")):
            if not effect.get_option("render_" + stage):
                nodepath.hide(self.tag_mgr.get_mask(stage))
            else:
                shader = effect.get_shader_obj(stage)
                if stage == "gbuffer":
                    nodepath.set_shader(shader, 25)
                else:
                    self.tag_mgr.apply_state(stage, nodepath, shader, str(effect.effect_id), 25 + 10 * i + sort)
                nodepath.show_through(self.tag_mgr.get_mask(stage))

        if effect.get_option("render_gbuffer") and effect.get_option("render_forward"):
            self.error(
                "You cannot render an object forward and deferred at the "
                "same time! Either use render_gbuffer or use render_forward, "
                "but not both."
            )

    def set_effect(self, nodepath, effect_src, options=None, sort=30):
        """ See _internal_set_effect. """
        args = (nodepath, effect_src, options, sort)
        self._applied_effects.append(args)
        self._internal_set_effect(*args)

    def add_environment_probe(self):
        """ Constructs a new environment probe and returns the handle, so that
        the probe can be modified. In case the env_probes plugin is not activated,
        this returns a dummy object which can be modified but has no impact. """
        if not self.plugin_mgr.is_plugin_enabled("env_probes"):
            self.warn("env_probes plugin is not loaded - cannot add environment probe")

            class DummyEnvironmentProbe(object):  # pylint: disable=too-few-public-methods
                def __getattr__(self, *args, **kwargs):
                    return lambda *args, **kwargs: None

            return DummyEnvironmentProbe()

        # Ugh ..
        from metadrive.render_pipeline.rpplugins.env_probes.environment_probe import EnvironmentProbe
        probe = EnvironmentProbe()
        self.plugin_mgr.instances["env_probes"].probe_mgr.add_probe(probe)
        return probe

    def prepare_scene(self, scene):
        """ Prepares a given scene, by converting panda lights to render pipeline
        lights. This also converts all empties with names starting with 'ENVPROBE'
        to environment probes. Conversion of blender to render pipeline lights
        is done by scaling their intensity by 100 to match lumens.

        Additionally, this finds all materials with the 'TRANSPARENT' shading
        model, and sets the proper effects on them to ensure they are rendered
        properly.

        This method also returns a dictionary with handles to all created
        objects, that is lights, environment probes, and transparent objects.
        This can be used to store them and process them later on, or delete
        them when a newer scene is loaded."""
        lights = []
        for light in scene.find_all_matches("**/+PointLight"):
            light_node = light.node()
            rp_light = PointLight()
            rp_light.pos = light.get_pos(Globals.base.render)
            rp_light.radius = light_node.max_distance
            rp_light.energy = 20.0 * light_node.color.w
            rp_light.color = light_node.color.xyz
            rp_light.casts_shadows = light_node.shadow_caster
            rp_light.shadow_map_resolution = light_node.shadow_buffer_size.x
            rp_light.inner_radius = 0.4

            self.add_light(rp_light)
            light.remove_node()
            lights.append(rp_light)

        for light in scene.find_all_matches("**/+Spotlight"):
            light_node = light.node()
            rp_light = SpotLight()
            rp_light.pos = light.get_pos(Globals.base.render)
            rp_light.radius = light_node.max_distance
            rp_light.energy = 20.0 * light_node.color.w
            rp_light.color = light_node.color.xyz
            rp_light.casts_shadows = light_node.shadow_caster
            rp_light.shadow_map_resolution = light_node.shadow_buffer_size.x
            rp_light.fov = light_node.exponent / math.pi * 180.0
            lpoint = light.get_mat(Globals.base.render).xform_vec((0, 0, -1))
            rp_light.direction = lpoint
            self.add_light(rp_light)
            light.remove_node()
            lights.append(rp_light)

        envprobes = []
        for np in scene.find_all_matches("**/ENVPROBE*"):
            probe = self.add_environment_probe()
            probe.set_mat(np.get_mat())
            probe.border_smoothness = 0.0001
            probe.parallax_correction = True
            np.remove_node()
            envprobes.append(probe)

        tristrips_warning_emitted = False
        transparent_objects = []
        for geom_np in scene.find_all_matches("**/+GeomNode"):
            geom_node = geom_np.node()
            geom_count = geom_node.get_num_geoms()
            for i in range(geom_count):
                state = geom_node.get_geom_state(i)
                geom = geom_node.get_geom(i)

                needs_conversion = False
                for prim in geom.get_primitives():
                    if isinstance(prim, GeomTristrips):
                        needs_conversion = True
                        if not tristrips_warning_emitted:
                            self.warn(
                                "At least one GeomNode (", geom_node.get_name(),
                                "and possible more..) contains tristrips."
                            )
                            self.warn("Due to a NVIDIA Driver bug, we have to convert them to triangles now.")
                            self.warn("Consider exporting your models with the Bam Exporter to avoid this.")
                            tristrips_warning_emitted = True
                            break

                if needs_conversion:
                    geom_node.modify_geom(i).decompose_in_place()

                if not state.has_attrib(MaterialAttrib):
                    self.warn("Geom", geom_node, "has no material! Please fix this.")
                    continue

                material = state.get_attrib(MaterialAttrib).get_material()
                shading_model = material.emission.x

                # SHADING_MODEL_TRANSPARENT
                if shading_model == 3:
                    if geom_count > 1:
                        self.error(
                            "Transparent materials must be on their own geom!\n"
                            "If you are exporting from blender, split them into\n"
                            "seperate meshes, then re-export your scene. The\n"
                            "problematic mesh is: " + geom_np.get_name()
                        )
                        continue
                    self.set_effect(
                        geom_np, "effects/default.yaml", {
                            "render_forward": True,
                            "render_gbuffer": False
                        }, 100
                    )

        return {"lights": lights, "envprobes": envprobes, "transparent_objects": transparent_objects}

    def _create_managers(self):
        """ Internal method to create all managers and instances. This also
        initializes the commonly used render stages, which are always required,
        independently of which plugins are enabled. """
        self.task_scheduler = TaskScheduler(self)
        self.tag_mgr = TagStateManager(Globals.base.cam)
        self.plugin_mgr = PluginManager(self)
        self.stage_mgr = StageManager(self)
        self.light_mgr = LightManager(self)
        self.daytime_mgr = DayTimeManager(self)
        self.ies_loader = IESProfileLoader(self)
        self.common_resources = CommonResources(self)
        self._init_common_stages()

    def _analyze_system(self):
        """ Prints information about the system used, including information
        about the used Panda3D build. Also checks if the Panda3D build is out
        of date. """
        self.debug(
            "Using Python {}.{} with architecture {}".format(
                sys.version_info.major, sys.version_info.minor, PandaSystem.get_platform()
            )
        )

        build_date = getattr(PandaSystem, 'build_date', None)
        if build_date:
            self.debug("Using Panda3D {} built on {}".format(PandaSystem.get_version_string(), build_date))
        else:
            self.debug("Using Panda3D {}".format(PandaSystem.get_version_string()))

        if PandaSystem.get_git_commit():
            self.debug("Using git commit {}".format(PandaSystem.get_git_commit()))
        else:
            self.debug("Using custom Panda3D build")
        if not self._check_version():
            self.fatal(
                "Your Panda3D version is outdated! Please update to the newest \n"
                "git version! Checkout https://github.com/panda3d/panda3d to "
                "compile panda from source, or get a recent buildbot build."
            )

    def _initialize_managers(self):
        """ Internal method to initialize all managers, after they have been
        created earlier in _create_managers. The creation and initialization
        is seperated due to the fact that plugins and various other subprocesses
        have to get initialized inbetween. """
        self.stage_mgr.setup()
        self.stage_mgr.reload_shaders()
        self.light_mgr.reload_shaders()
        self._init_bindings()
        self.light_mgr.init_shadows()

    def _init_debugger(self):
        """ Internal method to initialize the GUI-based debugger. In case debugging
        is disabled, this constructs a dummy debugger, which does nothing.
        The debugger itself handles the various onscreen components. """
        if self.settings["pipeline.display_debugger"]:
            self.debugger = Debugger(self)
        else:
            # Use an empty onscreen debugger in case the debugger is not
            # enabled, which defines all member functions as empty lambdas
            class EmptyDebugger(object):  # pylint: disable=too-few-public-methods
                def __getattr__(self, *args, **kwargs):
                    return lambda *args, **kwargs: None

            self.debugger = EmptyDebugger()  # pylint: disable=redefined-variable-type
            del EmptyDebugger

    def _init_globals(self):
        """ Inits all global bindings. This includes references to the global
        ShowBase instance, as well as the render resolution, the GUI font,
        and various global logging and output methods. """
        Globals.load(self._showbase)
        native_w, native_h = self._showbase.win.get_x_size(), self._showbase.win.get_y_size()
        Globals.native_resolution = LVecBase2i(native_w, native_h)
        self._last_window_dims = LVecBase2i(Globals.native_resolution)
        self._compute_render_resolution()
        RenderTarget.RT_OUTPUT_FUNC = lambda *args: RPObject.global_warn("RenderTarget", *args[1:])
        RenderTarget.USE_R11G11B10 = self.settings["pipeline.use_r11_g11_b10"]

    def _set_default_effect(self):
        """ Sets the default effect used for all objects if not overridden, this
        just calls set_effect with the default effect and options as parameters.
        This uses a very low sort, to make sure that overriding the default
        effect does not require a custom sort parameter to be passed. """
        self.set_effect(Globals.render, "effects/default.yaml", {}, -10)

    def _adjust_camera_settings(self):
        """ Sets the default camera settings, this includes the cameras
        near and far plane, as well as FoV. The reason for this is, that pandas
        default field of view is very small, and thus we increase it. """
        self._showbase.camLens.set_near_far(0.1, 70000)
        self._showbase.camLens.set_fov(40)

    def _compute_render_resolution(self):
        """ Computes the internally used render resolution. This might differ
        from the window dimensions in case a resolution scale is set. """
        scale_factor = self.settings["pipeline.resolution_scale"]
        w = int(float(Globals.native_resolution.x) * scale_factor)
        h = int(float(Globals.native_resolution.y) * scale_factor)
        # Make sure the resolution is a multiple of 4
        w, h = w - w % 4, h - h % 4
        self.debug("Render resolution is", w, "x", h)
        Globals.resolution = LVecBase2i(w, h)

    def _init_showbase(self, base):
        """ Inits the the given showbase object. This is part of an alternative
        method of initializing the showbase. In case base is None, a new
        ShowBase instance will be created and initialized. Otherwise base() is
        expected to either be an uninitialized ShowBase instance, or an
        initialized instance with pre_showbase_init() called inbefore. """
        if not base:
            self.pre_showbase_init()
            self._showbase = ShowBase()
        else:
            if not hasattr(base, "render"):
                self.pre_showbase_init()
                ShowBase.__init__(base)
            else:
                if not self._pre_showbase_initialized:
                    self.fatal(
                        "You constructed your own ShowBase object but you\n"
                        "did not call pre_show_base_init() on the render\n"
                        "pipeline object before! Checkout the 00-Loading the\n"
                        "pipeline sample to see how to initialize the RP."
                    )
            self._showbase = base

        # Now that we have a showbase and a window, we can print out driver info
        self.debug("Driver Version =", self._showbase.win.gsg.driver_version)
        self.debug("Driver Vendor =", self._showbase.win.gsg.driver_vendor)
        self.debug("Driver Renderer =", self._showbase.win.gsg.driver_renderer)

    def _init_bindings(self):
        """ Internal method to init the tasks and keybindings. This constructs
        the tasks to be run on a per-frame basis. """
        self._showbase.addTask(self._manager_update_task, "RP_UpdateManagers", sort=10)
        self._showbase.addTask(self._plugin_pre_render_update, "RP_Plugin_BeforeRender", sort=12)
        self._showbase.addTask(self._plugin_post_render_update, "RP_Plugin_AfterRender", sort=15)
        self._showbase.addTask(self._update_inputs_and_stages, "RP_UpdateInputsAndStages", sort=18)
        self._showbase.taskMgr.doMethodLater(0.5, self._clear_state_cache, "RP_ClearStateCache")
        self._showbase.accept("window-event", self._handle_window_event)

    def _handle_window_event(self, event):
        """ Checks for window events. This mainly handles incoming resizes,
        and calls the required handlers """
        self._showbase.windowEvent(event)
        window_dims = LVecBase2i(self._showbase.win.get_x_size(), self._showbase.win.get_y_size())
        if window_dims != self._last_window_dims and window_dims != Globals.native_resolution:
            self._last_window_dims = LVecBase2i(window_dims)

            # Ensure the dimensions are a multiple of 4, and if not, correct it
            if window_dims.x % 4 != 0 or window_dims.y % 4 != 0:
                self.debug("Correcting non-multiple of 4 window size:", window_dims)
                window_dims.x = window_dims.x - window_dims.x % 4
                window_dims.y = window_dims.y - window_dims.y % 4
                props = WindowProperties.size(window_dims.x, window_dims.y)
                self._showbase.win.request_properties(props)

            self.debug("Resizing to", window_dims.x, "x", window_dims.y)
            Globals.native_resolution = window_dims
            self._compute_render_resolution()
            self.light_mgr.compute_tile_size()
            self.stage_mgr.handle_window_resize()
            self.debugger.handle_window_resize()
            self.plugin_mgr.trigger_hook("window_resized")

    def _clear_state_cache(self, task=None):
        """ Task which repeatedly clears the state cache to avoid storing
        unused states. While running once a while, this task prevents over-polluting
        the state-cache with unused states. This complements Panda3D's internal
        state garbarge collector, which does a great job, but still cannot clear
        up all states. """
        task.delayTime = 2.0
        TransformState.clear_cache()
        RenderState.clear_cache()
        return task.again

    def _manager_update_task(self, task):
        """ Update task which gets called before the rendering, and updates
        all managers."""
        self.task_scheduler.step()
        self._listener.update()
        self.debugger.update()
        self.daytime_mgr.update()
        self.light_mgr.update()

        if Globals.clock.get_frame_count() >= 10 and self.loading_screen.fullscreen_node.hasParent():
            self.debug("Hiding loading screen after 10 pre-rendered frames.")
            self.loading_screen.remove()

        return task.cont

    def _update_inputs_and_stages(self, task):
        """ Updates the commonly used inputs each frame. This is a seperate
        task to be able view detailed performance information in pstats, since
        a lot of matrix calculations are involved here. """
        self.common_resources.update()
        self.stage_mgr.update()
        return task.cont

    def _plugin_pre_render_update(self, task):
        """ Update task which gets called before the rendering, and updates the
        plugins. This is a seperate task to split the work, and be able to do
        better performance analysis in pstats later on. """
        self.plugin_mgr.trigger_hook("pre_render_update")
        return task.cont

    def _plugin_post_render_update(self, task):
        """ Update task which gets called after the rendering, and should cleanup
        all unused states and objects. This also triggers the plugin post-render
        update hook. """
        self.plugin_mgr.trigger_hook("post_render_update")
        if self._first_frame is not None:
            duration = time.process_time() - self._first_frame
            self.debug("Took", round(duration, 3), "s until first frame")
            self._first_frame = None
        return task.cont

    def _create_common_defines(self):
        """ Creates commonly used defines for the shader configuration. """
        defines = self.stage_mgr.defines
        defines["CAMERA_NEAR"] = round(Globals.base.camLens.get_near(), 10)
        defines["CAMERA_FAR"] = round(Globals.base.camLens.get_far(), 10)

        # Work arround buggy nvidia driver, which expects arrays to be const
        if "NVIDIA 361.43" in self._showbase.win.gsg.get_driver_version():
            defines["CONST_ARRAY"] = "const"
        else:
            defines["CONST_ARRAY"] = ""

        # Provide driver vendor as a define
        vendor = self._showbase.win.gsg.get_driver_vendor().lower()
        defines["IS_NVIDIA"] = "nvidia" in vendor
        defines["IS_AMD"] = vendor.startswith("ati")
        defines["IS_INTEL"] = "intel" in vendor

        defines["REFERENCE_MODE"] = self.settings["pipeline.reference_mode"]
        self.light_mgr.init_defines()
        self.plugin_mgr.init_defines()

    def _create_default_skybox(self, size=40000):
        """ Returns the default skybox, with a scale of <size>, and all
        proper effects and shaders already applied. The skybox is already
        parented to render as well. """
        skybox = self.common_resources.load_default_skybox()
        skybox.set_scale(size)
        skybox.reparent_to(Globals.render)
        skybox.set_bin("unsorted", 10000)
        self.set_effect(
            skybox, "effects/skybox.yaml", {
                "render_shadow": False,
                "render_envmap": False,
                "render_voxelize": False,
                "alpha_testing": False,
                "normal_mapping": False,
                "parallax_mapping": False
            }, 1000
        )
        return skybox

    def _check_version(self):
        """ Internal method to check if the required Panda3D version is met. Returns
        True if the version is new enough, and False if the version is outdated. """
        from panda3d.core import Texture
        if not hasattr(Texture, "F_r16i"):
            return False
        return True

    def _init_common_stages(self):
        """ Inits the commonly used stages, which don't belong to any plugin,
        but yet are necessary and widely used. """
        add_stage = self.stage_mgr.add_stage
        self._ambient_stage = AmbientStage(self)
        add_stage(self._ambient_stage)
        self._gbuffer_stage = GBufferStage(self)
        add_stage(self._gbuffer_stage)
        self._final_stage = FinalStage(self)
        add_stage(self._final_stage)
        self._downscale_stage = DownscaleZStage(self)
        add_stage(self._downscale_stage)
        self._combine_velocity_stage = CombineVelocityStage(self)
        add_stage(self._combine_velocity_stage)

        # Add an upscale/downscale stage in case we render at a different resolution
        if abs(1 - self.settings["pipeline.resolution_scale"]) > 0.005:
            self._upscale_stage = UpscaleStage(self)
            add_stage(self._upscale_stage)

    def _get_serialized_material_name(self, material, index=0):
        """ Returns a serializable material name """
        return str(index) + "-" + (material.get_name().replace(" ", "").strip() or "unnamed")

    def export_materials(self, pth):
        """ Exports a list of all materials found in the current scene in a
        serialized format to the given path """

        with open(pth, "w") as handle:
            for i, material in enumerate(Globals.render.find_all_materials()):
                if not material.has_base_color() or not material.has_roughness() or not material.has_refractive_index():
                    print("Skipping non-pbr material:", material.name)
                    continue

                handle.write(
                    ("{} " * 11).format(
                        self._get_serialized_material_name(material, i),
                        material.base_color.x,
                        material.base_color.y,
                        material.base_color.z,
                        material.roughness,
                        material.refractive_index,
                        material.metallic,
                        material.emission.x,  # shading model
                        material.emission.y,  # normal strength
                        material.emission.z,  # arbitrary 0
                        material.emission.w,  # arbitrary 1
                    ) + "\n"
                )

    def update_serialized_material(self, data):
        """ Internal method to update a material from a given serialized material """
        name = data[0]

        for i, material in enumerate(Globals.render.find_all_materials()):
            if self._get_serialized_material_name(material, i) == name:
                material.set_base_color(Vec4(float(data[1]), float(data[2]), float(data[3]), 1.0))
                material.set_roughness(float(data[4]))
                material.set_refractive_index(float(data[5]))
                material.set_metallic(float(data[6]))
                material.set_emission(Vec4(
                    float(data[7]),
                    float(data[8]),
                    float(data[9]),
                    float(data[10]),
                ))

        RenderState.clear_cache()



# Author: Epihaius
# Date: 2019-06-21
#
# This module contains a couple of imports and a base class needed for the
# creation of all 3D primitives.

from panda3d.core import *
from math import pi, sin, cos, acos, atan2, sqrt
import array


class ModelMaker:
    @property
    def segments(self):
        return self._segments

    @segments.setter
    def segments(self, segments):
        self._segments = segments

    @property
    def inverted(self):
        return self._inverted

    @inverted.setter
    def inverted(self, inverted):
        self._inverted = inverted

    @property
    def vertex_color(self):
        return self._vertex_color

    @vertex_color.setter
    def vertex_color(self, vertex_color):
        self._vertex_color = vertex_color

    @property
    def has_uvs(self):
        return self._has_uvs

    @has_uvs.setter
    def has_uvs(self, has_uvs):
        self._has_uvs = has_uvs

    @property
    def tex_units(self):
        return self._tex_units

    @tex_units.setter
    def tex_units(self, tex_units):
        self._tex_units = tex_units

    @property
    def tex_offset(self):
        return self._tex_offset

    @tex_offset.setter
    def tex_offset(self, tex_offset):
        self._tex_offset = tex_offset

    @property
    def tex_rotation(self):
        return self._tex_rotation

    @tex_rotation.setter
    def tex_rotation(self, tex_rotation):
        self._tex_rotation = tex_rotation

    @property
    def tex_scale(self):
        return self._tex_scale

    @tex_scale.setter
    def tex_scale(self, tex_scale):
        self._tex_scale = tex_scale

    @property
    def vertex_ranges(self):
        """
        The ranges of vertex indices for each surface, as a dict of
        (start_index, end_index) tuples (empty tuple if the surface
        was not created), with end_index not included in the range.

        See derived class documentation for available surfaces.

        """
        return self._vert_ranges

    def __init__(
        self, segments, inverted, vertex_color, has_uvs, tex_units, tex_offset, tex_rotation, tex_scale, surface_ids
    ):
        """
        This class generates model primitives with the given parameters, common to
        all primitive types:

            segments (dict of ints):
                the number of subdivisions of each surface (listed in derived class
                documentation);
                default = None (use defaults for all surfaces);
            inverted (bool):
                whether or not the geometry should be rendered inside-out;
                default is False;
            vertex_color (tuple or None):
                the color applied to all vertices (if not specified, the vertex format
                of the created geometry will not allow any color data to be set);
                default = None;
            has_uvs (bool):
                whether or not the model should have texture coordinates;
                default is True;
            tex_units (dict of float tuples):
                the texture size (width, height) in object-space units for each
                surface (listed in derived class documentation);
                default = None;
            tex_offset, tex_rotation, tex_scale (dicts of float tuples):
                the 2D texture transforms in object-space for each surface
                (listed in derived class documentation):
                    tex_offset: (u offset, v offset);
                    tex_rotation: angle in degrees;
                    tex_scale: (u scale, v scale);
                default = None.

        """

        self._segments = segments
        self._inverted = inverted
        self._vertex_color = vertex_color
        self._has_uvs = has_uvs
        self._tex_units = tex_units
        self._tex_offset = tex_offset
        self._tex_rotation = tex_rotation
        self._tex_scale = tex_scale
        self._surface_ids = surface_ids
        self._vert_ranges = {s_id: () for s_id in surface_ids}

    def reset(self):

        self._segments = None
        self._inverted = False
        self._vertex_color = None
        self._has_uvs = True
        self._tex_units = None
        self._tex_offset = None
        self._tex_rotation = None
        self._tex_scale = None
        self._vert_ranges = {s_id: () for s_id in self._surface_ids}

    def _make_flat_shaded(self, indices, verts):

        points = [Point3(verts[i]["pos"]) for i in indices[:3]]
        normal = Plane(*points).get_normal()

        for i in indices:
            verts[i]["normal"] = normal

    def _average_normals(self, index1, index2, verts):

        normal = (verts[index1]["normal"] + verts[index2]["normal"]).normalized()
        verts[index1]["normal"] = normal
        verts[index2]["normal"] = normal

    def _get_tex_xform(self, surface_id):

        tex_offset = self._tex_offset
        tex_rotation = self._tex_rotation
        tex_scale = self._tex_scale

        has_tex_offset = tex_offset and surface_id in tex_offset
        has_tex_rot = tex_rotation and surface_id in tex_rotation
        has_tex_scale = tex_scale and surface_id in tex_scale
        has_tex_xform = has_tex_offset or has_tex_rot or has_tex_scale

        if has_tex_xform:

            mat = Mat3.ident_mat()

            if has_tex_scale:
                mat = mat * Mat3.scale_mat(*tex_scale[surface_id])
            if has_tex_rot:
                mat = mat * Mat3.rotate_mat(tex_rotation[surface_id])
            if has_tex_offset:
                mat = mat * Mat3.translate_mat(*tex_offset[surface_id])

            return mat



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division

from metadrive.render_pipeline.rplibs.six import iteritems
from direct.stdpy.file import open

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.util.shader_input_blocks import GroupedInputBlock
from metadrive.render_pipeline.rpcore.pluginbase.day_setting_types import ColorType


class DayTimeManager(RPObject):
    """ This manager handles all time of day settings, provides them as
    a input to all shaders, and stores which time it currently is """
    def __init__(self, pipeline):
        RPObject.__init__(self)
        self._pipeline = pipeline
        self._input_ubo = GroupedInputBlock("TimeOfDay")
        self._time = 0.5
        self._setting_handles = {}

    @property
    def time(self):
        """ Returns the current time of day as floating point number
        from 0 to 1, whereas 0 means 00:00 and 1 means 24:00 (=00:00) """
        return self._time

    @time.setter
    def time(self, day_time):
        """ Sets the current time of day as floating point number from
        0 to 1, whereas 0 means 00:00 and 1 means 24:00 (=00:00). Any
        number greater than 1 will be reduced to fit the 0 .. 1 range by
        doing time modulo 1.

        Alternatively a string in the format 'hh:mm' can be passed. """
        if isinstance(day_time, float):
            self._time = day_time % 1.0
        elif isinstance(day_time, str):
            parts = [int(i) for i in day_time.split(":")]
            self._time = (parts[0] * 60 + parts[1]) / (24 * 60)

    @property
    def formatted_time(self):
        """ Returns the current time as formatted string, e.g. 12:34 """
        total_minutes = int(self._time * 24 * 60)
        return "{:02d}:{:02d}".format(total_minutes // 60, total_minutes % 60)

    def load_settings(self):
        """ Loads all day time settings from the plugin manager and registers
        them to the used input buffer """
        for plugin_id, settings in iteritems(self._pipeline.plugin_mgr.day_settings):
            for setting, handle in iteritems(settings):
                setting_id = "{}.{}".format(plugin_id, setting)
                self._input_ubo.register_pta(setting_id, handle.glsl_type)
                self._setting_handles[setting_id] = handle
        self._pipeline.stage_mgr.input_blocks.append(self._input_ubo)

        # Generate UBO shader code
        shader_code = self._input_ubo.generate_shader_code()
        with open("/$$rptemp/$$daytime_config.inc.glsl", "w") as handle:
            handle.write(shader_code)

    def update(self):
        """ Internal update method which updates all day time settings """
        for setting_id, handle in iteritems(self._setting_handles):
            # XXX: Find a better interface for this. Without this fix, colors
            # are in the range 0 .. 255 in the shader.
            if isinstance(handle, ColorType):
                value = handle.get_value_at(self._time)
            else:
                value = handle.get_scaled_value_at(self._time)
            self._input_ubo.update_input(setting_id, value)



from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.component.vehicle.PID_controller import PIDController
from metadrive.envs.metadrive_env import MetaDriveEnv


class Target:
    def __init__(self, target_lateral, target_speed_km_h):
        self.lateral = target_lateral
        self.speed_km_h = target_speed_km_h

    def go_right(self):
        self.lateral += 0.25 if self.lateral < 0.625 else 0

    def go_left(self):
        self.lateral -= 0.25 if self.lateral > 0.125 else 0

    def faster(self):
        self.speed_km_h += 10

    def slower(self):
        self.speed_km_h -= 10


def test_navigation(vis=False):
    env = MetaDriveEnv(
        {
            "num_scenarios": 10,
            "traffic_density": 0.0,
            "use_render": vis,
            "start_seed": 5,
            "map_config": {
                BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_NUM,
                BaseMap.GENERATE_CONFIG: 7,
                BaseMap.LANE_WIDTH: 3.5,
                BaseMap.LANE_NUM: 3,
            }
        }
    )
    target = Target(0.375, 30)
    o, _ = env.reset()
    if vis:
        env.engine.accept('d', target.go_right)
        env.engine.accept('a', target.go_left)
        env.engine.accept('w', target.faster)
        env.engine.accept('s', target.slower)

    steering_controller = PIDController(1.6, 0.0008, 27.3)
    acc_controller = PIDController(0.1, 0.001, 0.3)

    steering_error = o[0] - target.lateral
    steering = steering_controller.get_result(steering_error)

    acc_error = env.vehicles[env.DEFAULT_AGENT].speed_km_h - target.speed_km_h
    acc = acc_controller.get_result(acc_error)
    for i in range(1, 1000000 if vis else 2000):
        o, r, tm, tc, info = env.step([-steering, acc])
        # calculate new action

        steering_error = o[0] - target.lateral
        steering = steering_controller.get_result(steering_error)

        t_speed = target.speed_km_h if abs(o[12] - 0.5) < 0.01 else target.speed_km_h - 10
        acc_error = env.vehicles[env.DEFAULT_AGENT].speed_km_h - t_speed
        acc = acc_controller.get_result(acc_error)
        if vis:
            if i < 700:
                env.render(
                    text={
                        "W": "Target speed +",
                        "S": "Target speed -",
                        "A": "Change to left lane",
                        "D": "Change to right lane"
                    }
                )
            if i == 500:
                env.engine.on_screen_message.data.clear()
            else:
                env.render()
        if tm or tc:
            # print("Reset")
            o, _ = env.reset()

            steering_controller.reset()
            steering_error = o[0] - target.lateral
            steering = steering_controller.get_result(steering_error, o[11])

            acc_controller.reset()
            acc_error = env.vehicles[env.DEFAULT_AGENT].speed_km_h - target.speed_km_h
            acc = acc_controller.get_result(acc_error)
    env.close()


if __name__ == "__main__":
    test_navigation(vis=True)



import logging
from typing import Union

from panda3d.core import NodePath

from metadrive.component.algorithm.blocks_prob_dist import PGBlockDistConfig
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.pgblock.pg_block import PGBlock
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.engine.core.physics_world import PhysicsWorld
from metadrive.utils import get_np_random, get_metadrive_class


class NextStep:
    back = 0
    forward = 1
    search_sibling = 3
    destruct_current = 4


class BigGenerateMethod:
    BLOCK_SEQUENCE = "block_sequence"
    BLOCK_NUM = "block_num"
    SINGLE_BLOCK = "single_block"


class BIG:
    MAX_TRIAL = 5

    def __init__(
        self,
        lane_num: int,
        lane_width: float,
        global_network: NodeRoadNetwork,
        render_node_path: NodePath,
        physics_world: PhysicsWorld,
        # block_type_version: str,
        exit_length=50,
        random_seed=None,
        block_dist_config=PGBlockDistConfig
    ):
        self.block_dist_config = block_dist_config
        self._block_sequence = None
        self.random_seed = random_seed
        self.np_random = get_np_random(random_seed)
        # Don't change this right now, since we need to make maps identical to old one
        self._lane_num = lane_num
        self._lane_width = lane_width
        self.block_num = None
        self._render_node_path = render_node_path
        self._physics_world = physics_world
        self._global_network = global_network
        self.blocks = []
        self._exit_length = exit_length
        first_block = FirstPGBlock(
            self._global_network,
            self._lane_width,
            self._lane_num,
            self._render_node_path,
            self._physics_world,
            length=self._exit_length
        )
        self.blocks.append(first_block)
        self.next_step = NextStep.forward
        # assert block_type_version in ["v1", "v2"]
        # self.block_type_version = block_type_version

    def generate(self, generate_method: str, parameter: Union[str, int]):
        """
        In order to embed it to the show_base loop, we implement BIG in a more complex way
        """
        if generate_method == BigGenerateMethod.BLOCK_NUM:
            assert isinstance(parameter, int), "When generating map by assigning block num, the parameter should be int"
            self.block_num = parameter + 1
        elif generate_method == BigGenerateMethod.BLOCK_SEQUENCE:
            assert isinstance(parameter, str), "When generating map from block sequence, the parameter should be a str"
            self.block_num = len(parameter) + 1
            self._block_sequence = FirstPGBlock.ID + parameter
        while True:
            if self.big_helper_func():
                break
        return self._global_network

    def big_helper_func(self):
        if len(self.blocks) >= self.block_num and self.next_step == NextStep.forward:
            return True
        if self.next_step == NextStep.forward:
            self._forward()
        elif self.next_step == NextStep.destruct_current:
            self._destruct_current()
        elif self.next_step == NextStep.search_sibling:
            self._search_sibling()
        elif self.next_step == NextStep.back:
            self._go_back()
        return False

    def sample_block(self) -> PGBlock:
        """
        Sample a random block type
        """
        if self._block_sequence is None:
            block_types = self.block_dist_config.all_blocks()
            block_probabilities = self.block_dist_config.block_probability()
            block_type = self.np_random.choice(block_types, p=block_probabilities)
            block_type = get_metadrive_class(block_type)
        else:
            type_id = self._block_sequence[len(self.blocks)]
            block_type = self.block_dist_config.get_block(type_id)

        socket = self.np_random.choice(self.blocks[-1].get_socket_indices())
        block = block_type(
            len(self.blocks),
            self.blocks[-1].get_socket(socket),
            self._global_network,
            self.np_random.randint(0, 10000),
            ignore_intersection_checking=False
        )
        return block

    def destruct(self, block):
        block.destruct_block(self._physics_world)

    def construct(self, block) -> bool:
        success = block.construct_block(self._render_node_path, self._physics_world)
        lane_num = max([len(socket.get_positive_lanes(self._global_network)) for socket in block._sockets.values()])
        if lane_num < self.block_dist_config.MIN_LANE_NUM or lane_num > self.block_dist_config.MAX_LANE_NUM:
            success = False
        return success

    def _forward(self):
        logging.debug("forward")
        block = self.sample_block()
        self.blocks.append(block)
        success = self.construct(block)
        self.next_step = NextStep.forward if success else NextStep.destruct_current

    def _go_back(self):
        logging.debug("back")

        to_delete_block = self.blocks.pop()
        to_delete_block.destroy()
        del to_delete_block

        last_block = self.blocks[-1]
        self.destruct(last_block)
        self.next_step = NextStep.search_sibling

    def _search_sibling(self):
        logging.debug("sibling")
        block = self.blocks[-1]
        if len(self.blocks) == 1:
            self.next_step = NextStep.forward
            return
        if block.number_of_sample_trial < self.MAX_TRIAL:
            success = self.construct(block)
            self.next_step = NextStep.forward if success else NextStep.destruct_current
        else:
            self.next_step = NextStep.back

    def _destruct_current(self):
        logging.debug("destruct")
        block = self.blocks[-1]
        self.destruct(block)
        self.next_step = NextStep.search_sibling if block.number_of_sample_trial < self.MAX_TRIAL else NextStep.back

    def __del__(self):
        logging.debug("Destroy Big")



from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.policy.idm_policy import IDMPolicy
from metadrive.envs.marl_envs.multi_agent_metadrive import MULTI_AGENT_METADRIVE_DEFAULT_CONFIG
MULTI_AGENT_METADRIVE_DEFAULT_CONFIG["force_seed_spawn_manager"] = True


def test_ma_bidirection_idm(render=False):
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0,
            "start_seed": 22,
            "manual_control": False,
            "use_render": render,
            "map": "yBY",
            "agent_policy": IDMPolicy,
        }
    )
    index = ('1y0_1_', '2B0_0_', 0)
    try:
        o, _ = env.reset()
        env.vehicle.set_velocity([1, 0.1], 10)
        # print(env.vehicle.speed)
        pass_test = False
        for s in range(1, 10000):
            o, r, tm, tc, info = env.step(env.action_space.sample())
            _, lat = env.vehicle.lane.local_coordinates(env.vehicle.position)
            if abs(lat) > env.vehicle.lane.width / 2 + 0.1 and len(env.vehicle.navigation.current_ref_lanes) == 1:
                raise ValueError("IDM can not pass bidirection block")
            if env.vehicle.lane.index == index and abs(lat) < 0.1:
                pass_test = True
            if (tm or tc) and info["arrive_dest"]:
                break
        assert pass_test
        env.close()
    finally:
        env.close()


if __name__ == "__main__":
    test_ma_bidirection_idm(True)



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


def get_root_node(node_tree, type_):
    for node in node_tree.nodes:
        if node.type == type_:
            if type_ == 'OUTPUT_MATERIAL':
                if not node.is_active_output:
                    continue
            return node


def get_from_node(node_tree, type, to_node, from_socket_name, to_socket_name):
    for link in node_tree.links:
        # if (link.from_node.type == type and link.to_node == to_node and
        #         link.from_socket.name == from_socket_name and
        #         link.to_socket.name == to_socket_name):
        #     return link.from_node

        # right side matched:
        # -> [To Socket] To Node
        if (link.to_node == to_node and link.to_socket.name == to_socket_name):
            # left side matched:
            # From Node [From Socket] ->
            if (link.from_node.type == type and link.from_socket.name == from_socket_name):
                return link.from_node

            # through mix shader:
            # Mix Shader [Shader] ->
            elif (link.from_node.type == 'MIX_SHADER' and link.from_socket.name == 'Shader'):
                mix = link.from_node
                for input_ in mix.inputs:
                    if input_.name == 'Shader':
                        node = get_from_node(node_tree, type, mix, from_socket_name, 'Shader')
                        if node:
                            return node

            # through math:
            # Math [Value] ->
            elif (link.from_node.type == 'MATH' and link.from_socket.name == 'Value'):
                math = link.from_node
                for input_ in math.inputs:
                    if input_.name == 'Value':
                        node = get_from_node(node_tree, type, math, from_socket_name, 'Value')
                        if node:
                            return node

            # through separate RGB:
            # Separate RGB [Value] ->
            elif (link.from_node.type == 'SEPRGB' and link.from_socket.name in ('R', 'G', 'B')):
                seprgb = link.from_node
                node = get_from_node(node_tree, type, seprgb, from_socket_name, 'Image')
                if node:
                    return node



import logging

from metadrive.utils import random_string


class Nameable:
    """
    Instance of this class will have a special name
    """
    def __init__(self, name=None):
        # ID for object
        self.name = random_string() if name is None else name
        self.id = self.name  # name = id

    @property
    def class_name(self):
        return self.__class__.__name__

    def __del__(self):
        try:
            str(self)
        except AttributeError:
            pass
        else:
            logging.debug("{} is destroyed".format(str(self)))

    def __repr__(self):
        return "{}".format(str(self))

    def __str__(self):
        return "{}, ID:{}".format(self.class_name, self.name)

    def rename(self, new_name):
        self.name = new_name
        self.id = self.name
